# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unit tests for the rv session file adapter"""

import unittest

import opentimelineio as otio

MODULE = otio.adapters.from_name('burnins').module()
SAMPLE_DATA = """{
    "OTIO_SCHEMA": "Timeline.1",
    "metadata": {
        "burnins": {
            "overwrite": true,
            "burnins": [
                {
                    "text": "Top Center",
                    "align": "top_centered",
                    "font": "/System/Library/Fonts/Menlo.ttc",
                    "font_size": 48,
                    "function": "text"
                },
                {
                    "align": "top_left",
                    "x_offset": 75,
                    "font": "/System/Library/Fonts/Menlo.ttc",
                    "frame_offset": 101,
                    "font_size": 48,
                    "function": "frame_number"
                }
            ],
            "streams": [
                {
                    "codec_type": "video",
                    "codec_name": "h264",
                    "width": 1920,
                    "height": 1080,
                    "r_frame_rate": "30/1",
                    "start_time": "0.000000",
                    "duration": "20.000000"
                }
            ]
        }
    },
    "name": "TEST.MOV",
    "tracks": {
        "OTIO_SCHEMA": "Stack.1",
        "children": [
            {
                "OTIO_SCHEMA": "Track.1",
                "children": [
                    {
                        "OTIO_SCHEMA": "Clip.1",
                        "effects": [],
                        "markers": [],
                        "media_reference": {
                            "OTIO_SCHEMA": "ExternalReference.1",
                            "available_range": {
                                "OTIO_SCHEMA": "TimeRange.1",
                                "duration": {
                                    "OTIO_SCHEMA": "RationalTime.1",
                                    "rate": 30.0,
                                    "value": 600.0
                                },
                                "start_time": {
                                    "OTIO_SCHEMA": "RationalTime.1",
                                    "rate": 30.0,
                                    "value": 0.0
                                }
                            },
                            "metadata": {},
                            "name": null,
                            "target_url": "file://TEST.MOV"
                        },
                        "metadata": {},
                        "name": "TEST.MOV",
                        "source_range": null
                    }
                ],
                "effects": [],
                "kind": "Video",
                "markers": [],
                "metadata": {},
                "name": "TEST.MOV",
                "source_range": null
            }
        ],
        "effects": [],
        "markers": [],
        "metadata": {},
        "name": "tracks",
        "source_range": null
    }
}"""
WITH_BG = ('ffmpeg -loglevel panic -i TEST.MOV -vf "drawtext=text='
           '\'Top Center\':x=w/2-tw/2:y=0:fontcolor=white@1.0:fontsize'
           '=48:fontfile=\'/System/Library/Fonts/Menlo.ttc\':box=1:boxbord'
           'erw=5:boxcolor=black@1.0,drawtext=text=\''
           r'%{eif\:n+101\:d}'
           '\':x=75:y=0:fontcolor=white@1.0:fontsize=48:fontfile=\'/Syst'
           'em/Library/Fonts/Menlo.ttc\':box=1:boxborderw=5:boxcolor=bla'
           'ck@1.0" TEST.MOV')

WITHOUT_BG = ('ffmpeg -loglevel panic -i TEST.MOV -vf "drawtext=text='
              '\'Top Center\':x=w/2-tw/2:y=0:fontcolor=white@1.0:fontsize'
              '=48:fontfile=\'/System/Library/Fonts/Menlo.ttc\','
              'drawtext=text=\''
              r'%{eif\:n+101\:d}'
              '\':x=75:y=0:fontcolor=white@1.0:fontsize=48:fontfile=\'/System'
              '/Library/Fonts/Menlo.ttc\'" TEST.MOV')
TIMECODE = ('ffmpeg -loglevel panic -i TEST.MOV -vf "drawtext=timecode='
            '\'Top Center\':timecode_rate=24.00:x=w/2-tw/2:y=0:fontcolor='
            'white@1.0:fontsize=48:fontfile=\'/System/Library/Fonts/Menlo.'
            'ttc\':box=1:boxborderw=5:boxcolor=black@1.0,drawtext=timecode='
            r"'00\:00\:00\:00':timecode_rate=24.00:x=75:y=0:fontcolor="
            'white@1.0:fontsize=48:fontfile=\'/System/Library/Fonts/Menlo.'
            'ttc\':box=1:boxborderw=5:boxcolor=black@1.0" TEST.MOV')


try:
    import PIL # noqa
    from PIL.Image import core as imaging # noqa
    could_import_pillow = True
except (ImportError, SyntaxError):
    could_import_pillow = False


@unittest.skipIf(
    not could_import_pillow,
    "Pillow Required for burnin unit tests. see:"
    " https://python-pillow.org/"
)
class FFMPEGBurninsTest(unittest.TestCase):
    """Test Cases for FFMPEG Burnins"""

    def test_burnins_with_background(self):
        """
        Tests creating burnins with a background (box)
        """
        timeline = otio.adapters.read_from_string(SAMPLE_DATA, "otio_json")
        burnins = MODULE.build_burnins(timeline)
        self.assertEqual(len(burnins), 1)
        command = burnins[-1].command(burnins[-1].otio_media)
        self.assertEqual(command, WITH_BG)

    def test_burnins_without_background(self):
        """
        Tests creating burnins without a background (box)
        """
        timeline = otio.adapters.read_from_string(SAMPLE_DATA, "otio_json")
        for each in timeline.metadata['burnins']['burnins']:
            each['bg_color'] = None
        burnins = MODULE.build_burnins(timeline)
        self.assertEqual(len(burnins), 1)
        command = burnins[-1].command(burnins[-1].otio_media)
        self.assertEqual(command, WITHOUT_BG)

    def test_burnins_with_timecode(self):
        """
        Tests creating burnins with an animated timecode
        """
        timeline = otio.adapters.read_from_string(SAMPLE_DATA, "otio_json")
        for each in timeline.metadata['burnins']['burnins']:
            each['function'] = 'timecode'
            each['frame_offset'] = 0
            each['fps'] = 24
        burnins = MODULE.build_burnins(timeline)
        self.assertEqual(len(burnins), 1)
        command = burnins[-1].command(burnins[-1].otio_media)
        self.assertEqual(command, TIMECODE)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test harness for Composable."""

import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils


class ComposableTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    def test_constructor(self):
        seqi = otio.core.Composable(
            name="test",
            metadata={"foo": "bar"}
        )
        self.assertEqual(seqi.name, "test")
        self.assertEqual(seqi.metadata, {'foo': 'bar'})

    def test_serialize(self):
        seqi = otio.core.Composable(
            name="test",
            metadata={"foo": "bar"}
        )
        encoded = otio.adapters.otio_json.write_to_string(seqi)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(seqi, decoded)

    def test_stringify(self):
        seqi = otio.core.Composable()
        self.assertMultiLineEqual(
            str(seqi),
            "Composable("
            "{}, "
            "{}"
            ")".format(
                str(seqi.name),
                str(seqi.metadata),
            )
        )

        self.assertMultiLineEqual(
            repr(seqi),
            "otio.core.Composable("
            "name={}, "
            "metadata={}"
            ")".format(
                repr(seqi.name),
                repr(seqi.metadata),
            )
        )

    def test_metadata(self):
        seqi = otio.core.Composable()
        seqi.metadata["foo"] = "bar"
        encoded = otio.adapters.otio_json.write_to_string(seqi)
        decoded = otio.adapters.otio_json.read_from_string(encoded)
        self.assertIsOTIOEquivalentTo(seqi, decoded)
        self.assertEqual(decoded.metadata["foo"], seqi.metadata["foo"])


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test final cut pro xml."""

# python
import json
import os
import tempfile
import unittest
from xml.etree import cElementTree

from opentimelineio import (
    adapters,
    opentime,
    schema,
    test_utils,
)

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
FCP7_XML_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "premiere_example.xml")
SIMPLE_XML_PATH = os.path.join(SAMPLE_DATA_DIR, "sample_just_track.xml")
EMPTY_ELEMENT_XML_PATH = os.path.join(SAMPLE_DATA_DIR, "empty_name_tags.xml")
HIERO_XML_PATH = os.path.join(SAMPLE_DATA_DIR, "hiero_xml_export.xml")
FILTER_XML_EXAMPLE_PATH = os.path.join(
    SAMPLE_DATA_DIR, "premiere_example_filter.xml"
)
FILTER_JSON_EXAMPLE_PATH = os.path.join(
    SAMPLE_DATA_DIR, "premiere_example_filter.json"
)
GENERATOR_XML_EXAMPLE_PATH = os.path.join(
    SAMPLE_DATA_DIR, "premiere_generators.xml"
)


class TestFcp7XmlUtilities(unittest.TestCase, test_utils.OTIOAssertions):
    adapter = adapters.from_name('fcp_xml').module()

    def test_xml_tree_to_dict(self):
        self.maxDiff = None

        with open(FILTER_JSON_EXAMPLE_PATH) as f:
            ref_dict = json.load(f)

        tree = cElementTree.parse(FILTER_XML_EXAMPLE_PATH)
        filter_element = tree.getroot()
        xml_dict = self.adapter._xml_tree_to_dict(filter_element)

        self.assertEqual(xml_dict, ref_dict)

        out_xml = self.adapter._dict_to_xml_tree(xml_dict, "filter")
        out_xml_string = self.adapter._make_pretty_string(out_xml)

        with open(FILTER_XML_EXAMPLE_PATH) as f:
            orig_xml_string = f.read()

        self.assertEqual(out_xml_string.strip(), orig_xml_string.strip())

        # validate empty tag handling
        empty_element = cElementTree.fromstring(
            "<top><empty/></top>"
        )
        empty_element_dict = self.adapter._xml_tree_to_dict(empty_element)
        self.assertIsNone(empty_element_dict["empty"])

        empty_xml = self.adapter._dict_to_xml_tree(empty_element_dict, "top")
        self.assertIsNone(empty_xml.find("empty").text)

        roundtrip_dict = self.adapter._xml_tree_to_dict(empty_xml)
        self.assertEqual(empty_element_dict, roundtrip_dict)

    def test_bool_value(self):
        truthy_element = cElementTree.fromstring("<ntsc>TRUE</ntsc>")
        self.assertTrue(self.adapter._bool_value(truthy_element))

        falsy_element = cElementTree.fromstring("<ntsc>FALSE</ntsc>")
        self.assertFalse(self.adapter._bool_value(falsy_element))

    def test_backreference_for_id(self):
        item1 = schema.Clip(name="clip1")
        item1_prime = schema.Clip(name="clip1")
        item2 = schema.Clip(name="clip2")

        br_map = {}
        item1_id, item1_is_new = self.adapter._backreference_for_item(
            item1, "clipitem", br_map
        )
        self.assertEqual(item1_id, "clipitem-1")
        self.assertTrue(item1_is_new)

        item2_id, item2_is_new = self.adapter._backreference_for_item(
            item2, "clipitem", br_map
        )
        self.assertEqual(item2_id, "clipitem-2")
        self.assertTrue(item2_is_new)

        (
            item1_prime_id, item1_prime_is_new
        ) = self.adapter._backreference_for_item(
            item1_prime, "clipitem", br_map
        )
        self.assertEqual(item1_prime_id, "clipitem-1")
        self.assertFalse(item1_prime_is_new)

    def test_backreference_for_id_preserved(self):
        item1 = schema.Clip(
            name="clip23",
            metadata={"fcp_xml": {"@id": "clipitem-23"}},
        )
        item2 = schema.Clip(name="clip2")
        conflicting_item = schema.Clip(
            name="conflicting_clip",
            metadata={"fcp_xml": {"@id": "clipitem-1"}},
        )

        # prepopulate the backref map with some ids
        br_map = {
            "clipitem": {
                "bogus_hash": 1, "bogus_hash_2": 2, "bogus_hash_3": 3
            }
        }

        # Make sure the id is preserved
        item1_id, item1_is_new = self.adapter._backreference_for_item(
            item1, "clipitem", br_map
        )
        self.assertEqual(item1_id, "clipitem-23")
        self.assertTrue(item1_is_new)

        # Make sure the next item continues to fill in
        item2_id, item2_is_new = self.adapter._backreference_for_item(
            item2, "clipitem", br_map
        )
        self.assertEqual(item2_id, "clipitem-4")
        self.assertTrue(item2_is_new)

        # Make sure confilcting clips don't stomp existing ones
        item3_id, item3_is_new = self.adapter._backreference_for_item(
            conflicting_item, "clipitem", br_map
        )
        self.assertEqual(item3_id, "clipitem-5")
        self.assertTrue(item3_is_new)

    def test_name_from_element(self):
        sequence_element = cElementTree.fromstring(
            """
            <sequence>
                <name>My Sequence</name>
            </sequence>
            """
        )
        name = self.adapter._name_from_element(sequence_element)
        self.assertEqual(name, "My Sequence")

        empty_element = cElementTree.fromstring("<sequence></sequence>")
        empty_name = self.adapter._name_from_element(empty_element)
        self.assertEqual(empty_name, "")

        empty_name_element = cElementTree.fromstring(
            "<sequence><name></name></sequence>"
        )
        empty_name_2 = self.adapter._name_from_element(empty_name_element)
        self.assertEqual(empty_name_2, "")

    def test_rate_for_element_ntsc_conversion_23976(self):
        rate_element = cElementTree.fromstring(
            """
            <clipitem>
                <rate>
                    <timebase>24</timebase>
                    <ntsc>TRUE</ntsc>
                </rate>
            </clipitem>
            """
        )
        rate = self.adapter._rate_from_context(
            self.adapter._Context(rate_element)
        )

        self.assertEqual(rate, (24000 / 1001.0))

    def test_rate_for_element_ntsc_conversion_24(self):
        rate_element = cElementTree.fromstring(
            """
            <clipitem>
                <rate>
                    <timebase>24</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
            </clipitem>
            """
        )
        rate = self.adapter._rate_from_context(
            self.adapter._Context(rate_element)
        )

        self.assertEqual(rate, 24)

    def test_rate_for_element_ntsc_conversion_2997(self):
        rate_element = cElementTree.fromstring(
            """
            <clipitem>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>TRUE</ntsc>
                </rate>
            </clipitem>
            """
        )
        rate = self.adapter._rate_from_context(
            self.adapter._Context(rate_element)
        )

        self.assertEqual(rate, (30000 / 1001.0))

    def test_rate_for_element_ntsc_conversion_30(self):
        rate_element = cElementTree.fromstring(
            """
            <clipitem>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
            </clipitem>
            """
        )
        rate = self.adapter._rate_from_context(
            self.adapter._Context(rate_element)
        )

        self.assertEqual(rate, 30)

    def test_rate_for_element_no_ntsc(self):
        rate_element = cElementTree.fromstring(
            """
            <clipitem>
                <rate>
                    <timebase>30</timebase>
                </rate>
            </clipitem>
            """
        )
        rate = self.adapter._rate_from_context(
            self.adapter._Context(rate_element)
        )

        self.assertEqual(rate, 30)

    def test_rate_from_context(self):
        sequence_elem = cElementTree.fromstring(
            """
            <sequence>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>TRUE</ntsc>
                </rate>
            </sequence>
            """
        )

        # Fetch rate from one level of context
        sequence_context = self.adapter._Context(sequence_elem)
        sequence_rate = self.adapter._rate_from_context(sequence_context)
        self.assertEqual(sequence_rate, (30000 / 1001.0))

        track_elem = cElementTree.fromstring(
            """
            <track>
              <rate>
                <timebase>24</timebase>
                <ntsc>TRUE</ntsc>
              </rate>
            </track>
            """
        )

        # make sure pushing a context with a new rate overrides the old rate
        track_context = sequence_context.context_pushing_element(track_elem)
        track_noinherit_rate = self.adapter._rate_from_context(track_context)
        self.assertEqual(track_noinherit_rate, (24000 / 1001.0))

        clip_norate_elem = cElementTree.fromstring(
            """
            <clipitem>
            <name>Just soeme clip</name>
            </clipitem>
            """
        )

        # Make sure pushing a context element with no rate inherits the next
        # level up
        clip_context = track_context.context_pushing_element(clip_norate_elem)
        clip_inherit_rate = self.adapter._rate_from_context(clip_context)
        self.assertEqual(clip_inherit_rate, (24000 / 1001.0))

    def test_time_from_timecode_element(self):
        tc_element = cElementTree.fromstring(
            """
            <timecode>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
                <string>01:00:00:00</string>
                <frame>108000</frame>
                <displayformat>NDF</displayformat>
            </timecode>
            """
        )
        time = self.adapter._time_from_timecode_element(tc_element)

        self.assertEqual(time, opentime.RationalTime(108000, 30))

    def test_time_from_timecode_element_drop_frame(self):
        tc_element = cElementTree.fromstring(
            """
            <timecode>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>TRUE</ntsc>
                </rate>
                <string>10:03:00;05</string>
                <frame>1084319</frame>
                <displayformat>DF</displayformat>
            </timecode>
            """
        )
        time = self.adapter._time_from_timecode_element(tc_element)

        self.assertEqual(
            time, opentime.RationalTime(1084319, (30000 / 1001.0))
        )

    def test_time_from_timecode_element_ntsc_non_drop_frame(self):
        tc_element = cElementTree.fromstring(
            """
            <timecode>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>TRUE</ntsc>
                </rate>
                <string>00:59:56:12</string>
                <displayformat>NDF</displayformat>
            </timecode>
            """
        )

        time = self.adapter._time_from_timecode_element(tc_element)
        self.assertEqual(
            time, opentime.RationalTime(107892, (30000 / 1001.0))
        )

    def test_time_from_timecode_element_implicit_ntsc(self):
        clipitem_element = cElementTree.fromstring(
            """
            <clipitem>
              <duration>767</duration>
              <rate>
                <ntsc>TRUE</ntsc>
                <timebase>24</timebase>
              </rate>
              <in>447</in>
              <out>477</out>
              <start>264</start>
              <end>294</end>
              <file>
                <rate>
                  <timebase>24</timebase>
                  <ntsc>TRUE</ntsc>
                </rate>
                <duration>767</duration>
                <timecode>
                  <rate>
                    <timebase>24</timebase>
                  </rate>
                  <string>14:11:44:09</string>
                  <frame>1226505</frame>
                  <displayformat>NDF</displayformat>
                  <source>source</source>
                </timecode>
              </file>
            </clipitem>
            """
        )
        context = self.adapter._Context(clipitem_element)
        timecode_element = clipitem_element.find("./file/timecode")
        time = self.adapter._time_from_timecode_element(
            timecode_element, context
        )
        self.assertEqual(time, opentime.RationalTime(1226505, 24000.0 / 1001))

    def test_track_kind_from_element(self):
        video_element = cElementTree.fromstring("<video/>")
        video_kind = self.adapter._track_kind_from_element(video_element)
        self.assertEqual(video_kind, schema.TrackKind.Video)

        audio_element = cElementTree.fromstring("<audio/>")
        audio_kind = self.adapter._track_kind_from_element(audio_element)
        self.assertEqual(audio_kind, schema.TrackKind.Audio)

        invalid_element = cElementTree.fromstring("<smell/>")
        with self.assertRaises(ValueError):
            self.adapter._track_kind_from_element(invalid_element)

    def test_transition_cut_point(self):
        transition_element = cElementTree.fromstring(
            """
            <transitionitem>
                <start>538</start>
                <end>557</end>
                <alignment>end-black</alignment>
                <cutPointTicks>160876800000</cutPointTicks>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
                <effect>
                    <name>Cross Dissolve</name>
                    <effectid>Cross Dissolve</effectid>
                    <effectcategory>Dissolve</effectcategory>
                    <effecttype>transition</effecttype>
                    <mediatype>video</mediatype>
                    <wipecode>0</wipecode>
                    <wipeaccuracy>100</wipeaccuracy>
                    <startratio>0</startratio>
                    <endratio>1</endratio>
                    <reverse>FALSE</reverse>
                </effect>
            </transitionitem>
            """
        )
        alignment_element = transition_element.find("./alignment")

        track_element = cElementTree.fromstring(
            """
            <track>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
            </track>
            """
        )
        context = self.adapter._Context(track_element)

        cut_point = self.adapter._transition_cut_point(
            transition_element, context
        )
        self.assertEqual(cut_point, opentime.RationalTime(557, 30))

        alignment_element.text = "end-black"
        cut_point = self.adapter._transition_cut_point(
            transition_element, context
        )
        self.assertEqual(cut_point, opentime.RationalTime(557, 30))

        for alignment in ("start", "start-black"):
            alignment_element.text = alignment
            cut_point = self.adapter._transition_cut_point(
                transition_element, context
            )
            self.assertEqual(cut_point, opentime.RationalTime(538, 30))

        # TODO: Mathematically, this cut point falls at 547.5, is the rounding
        #       down behavior "correct"?
        alignment_element.text = "center"
        cut_point = self.adapter._transition_cut_point(
            transition_element, context
        )
        self.assertEqual(cut_point, opentime.RationalTime(547, 30))


class TestFcp7XmlElements(unittest.TestCase, test_utils.OTIOAssertions):
    """ Tests for isolated element parsers. """
    adapter = adapters.from_name('fcp_xml').module()

    def test_timeline_for_sequence(self):
        tree = cElementTree.parse(FCP7_XML_EXAMPLE_PATH)

        # Get the test sequence and pare out the track definitions to keep this
        # test simple.
        seq_elem = tree.find("sequence")
        seq_elem.find("./media").clear()
        seq_elem.find("./timecode/string").text = "01:00:00:00"
        seq_elem.find("./timecode/frame").text = "108000"

        parser = self.adapter.FCP7XMLParser(tree)
        context = self.adapter._Context()
        timeline = parser.timeline_for_sequence(seq_elem, context)

        # Spot-check the sequence
        self.assertEqual(timeline.name, "sc01_sh010_layerA")
        self.assertEqual(
            timeline.global_start_time, opentime.RationalTime(108000, 30)
        )

        # Spot check that metadata translated with a tag and a property
        adapter_metadata = timeline.metadata["fcp_xml"]
        self.assertEqual(
            adapter_metadata["labels"]["label2"], "Forest"
        )
        self.assertEqual(
            adapter_metadata["@MZ.Sequence.VideoTimeDisplayFormat"], "104"
        )

        # make sure the media and name tags were not included in the metadata
        for k in {"name", "media"}:
            with self.assertRaises(KeyError):
                adapter_metadata[k]

    def test_marker_for_element(self):
        marker_element = cElementTree.fromstring(
            """
            <marker>
              <comment>so, this happened</comment>
              <name>My MArker 1</name>
              <in>113</in>
              <out>-1</out>
            </marker>
            """
        )

        marker = self.adapter.marker_for_element(marker_element, 30)

        self.assertEqual(marker.name, "My MArker 1")
        self.assertEqual(
            marker.marked_range,
            opentime.TimeRange(
                start_time=opentime.RationalTime(113, 30),
                duration=opentime.RationalTime(0, 30),
            )
        )
        self.assertEqual(
            marker.metadata["fcp_xml"]["comment"], "so, this happened"
        )
        with self.assertRaises(KeyError):
            marker.metadata["fcp_xml"]["name"]

    def test_markers_from_element(self):
        sequence_element = cElementTree.fromstring(
            """
            <sequence>
              <rate>
                <timebase>30</timebase>
                <ntsc>FALSE</ntsc>
              </rate>
              <marker>
                <comment>so, this happened</comment>
                <name>My MArker 1</name>
                <in>113</in>
                <out>-1</out>
              </marker>
              <marker>
                <comment>fsfsfs</comment>
                <name>dsf</name>
                <in>492</in>
                <out>-1</out>
              </marker>
              <marker>
                <comment/>
                <name/>
                <in>298</in>
                <out>-1</out>
              </marker>
              <labels>
                <label2>Forest</label2>
              </labels>
            </sequence>
            """
        )
        markers = self.adapter.markers_from_element(sequence_element)

        # Note that "None" --
        expected_names = ["My MArker 1", "dsf", ""]
        self.assertEqual([m.name for m in markers], expected_names)

    def test_stack_from_element(self):
        tree = cElementTree.parse(FCP7_XML_EXAMPLE_PATH)

        # Get the test sequence and pare out the track definitions to keep this
        # test simple.
        media_elem = tree.find("./sequence/media")

        parser = self.adapter.FCP7XMLParser(tree)
        context = self.adapter._Context(tree.find("./sequence"))
        tracks = parser.stack_for_element(media_elem, context)

        self.assertEqual(len(tracks), 8)

        audio_tracks = [
            t for t in tracks if t.kind == schema.TrackKind.Audio
        ]
        self.assertEqual(len(audio_tracks), 4)

        video_tracks = [
            t for t in tracks if t.kind == schema.TrackKind.Video
        ]
        self.assertEqual(len(video_tracks), 4)

    def test_track_for_element(self):
        tree = cElementTree.parse(FCP7_XML_EXAMPLE_PATH)

        sequence_elem = tree.find("./sequence[1]")
        context = self.adapter._Context(sequence_elem)
        # The track with "clipitem-2" is a decent and relatively complex
        # test case
        track_elem = sequence_elem.find(".//clipitem[@id='clipitem-2']/..")

        # Make a parser and prime the id cache by parsing the file
        parser = self.adapter.FCP7XMLParser(tree)
        parser.timeline_for_sequence(sequence_elem, self.adapter._Context())

        track = parser.track_for_element(
            track_elem, schema.TrackKind.Video, context
        )

        expected_instance_types = [
            schema.Gap,
            schema.Clip,
            schema.Gap,
            schema.Clip,
            schema.Clip,
            schema.Transition,
            schema.Gap,
            schema.Stack,
        ]
        track_item_types = [i.__class__ for i in track]
        self.assertEqual(track_item_types, expected_instance_types)
        self.assertEqual(len(track), 8)

    def test_media_reference_from_element(self):
        file_element = cElementTree.fromstring(
            """
            <file id="file-3">
              <name>sc01_sh030_anim.mov</name>
              <pathurl>file:///Scratch/media/sc01_sh030_anim.2.mov</pathurl>
              <rate>
                <timebase>30</timebase>
                <ntsc>FALSE</ntsc>
              </rate>
              <duration>400</duration>
              <timecode>
                <rate>
                  <timebase>30</timebase>
                  <ntsc>FALSE</ntsc>
                </rate>
                <string>01:00:00:00</string>
                <frame>108000</frame>
                <displayformat>NDF</displayformat>
                <reel>
                  <name/>
                </reel>
              </timecode>
              <media>
                <video>
                  <samplecharacteristics>
                    <rate>
                      <timebase>30</timebase>
                      <ntsc>FALSE</ntsc>
                    </rate>
                    <width>1280</width>
                    <height>720</height>
                    <anamorphic>FALSE</anamorphic>
                    <pixelaspectratio>square</pixelaspectratio>
                    <fielddominance>none</fielddominance>
                  </samplecharacteristics>
                </video>
                <audio>
                  <samplecharacteristics>
                    <depth>16</depth>
                    <samplerate>48000</samplerate>
                  </samplecharacteristics>
                  <channelcount>2</channelcount>
                </audio>
              </media>
            </file>
            """
        )

        parser = self.adapter.FCP7XMLParser(file_element)
        context = self.adapter._Context()
        ref = parser.media_reference_for_file_element(
            file_element,
            context=context,
        )

        self.assertEqual(
            ref.target_url, "file:///Scratch/media/sc01_sh030_anim.2.mov"
        )
        self.assertEqual(ref.name, "sc01_sh030_anim.mov")
        self.assertEqual(
            ref.available_range,
            opentime.TimeRange(
                start_time=opentime.RationalTime(108000, 30),
                duration=opentime.RationalTime(400, 30),
            )
        )

        # Spot-check a metadata field
        video_metadata = ref.metadata["fcp_xml"]["media"]["video"]
        self.assertEqual(
            video_metadata["samplecharacteristics"]["height"], "720"
        )

    def test_missing_media_reference_from_element(self):
        file_element = cElementTree.fromstring(
            """
            <file id="101_021_0030_FG01">
              <name>101_021_0030_FG01</name>
              <duration>155</duration>
              <rate>
                <ntsc>FALSE</ntsc>
                <timebase>24</timebase>
              </rate>
              <timecode>
                <rate>
                  <ntsc>FALSE</ntsc>
                  <timebase>24</timebase>
                </rate>
                <frame>1308828</frame>
                <displayformat>NDF</displayformat>
                <string>15:08:54:12</string>
                <reel>
                  <name>A173C021_181204_R207</name>
                </reel>
              </timecode>
            </file>
            """
        )

        parser = self.adapter.FCP7XMLParser(file_element)
        context = self.adapter._Context()
        ref = parser.media_reference_for_file_element(
            file_element,
            context=context,
        )

        self.assertTrue(isinstance(ref, schema.MissingReference))
        self.assertEqual(ref.name, "101_021_0030_FG01")
        self.assertEqual(
            ref.available_range,
            opentime.TimeRange(
                start_time=opentime.RationalTime(1308828, 24),
                duration=opentime.RationalTime(155, 24),
            )
        )

        # Spot-check a metadata field
        reelname = ref.metadata["fcp_xml"]["timecode"]["reel"]["name"]
        self.assertEqual(reelname, "A173C021_181204_R207")

    def test_clip_for_element(self):
        tree = cElementTree.parse(FCP7_XML_EXAMPLE_PATH)

        # Use clipitem-3 because it's self-contained and doesn't reference
        # other elements
        sequence_elem = tree.find(".//clipitem[@id='clipitem-3']/../../../..")
        clip_elem = tree.find(".//clipitem[@id='clipitem-3']")
        context = self.adapter._Context(sequence_elem)

        # Make a parser
        parser = self.adapter.FCP7XMLParser(tree)

        clip, time_range = parser.item_and_timing_for_element(
            clip_elem,
            head_transition=None,
            tail_transition=None,
            context=context,
        )

        self.assertEqual(clip.name, "sc01_sh020_anim.mov")

        expected_range = opentime.TimeRange(
            start_time=opentime.RationalTime(165, 30),
            duration=opentime.RationalTime(157, 30),
        )
        self.assertEqual(time_range, expected_range)

        expected_range = opentime.TimeRange(
            start_time=opentime.RationalTime(0, 30),
            duration=opentime.RationalTime(157, 30),
        )
        self.assertEqual(clip.source_range, expected_range)

    def test_generator_for_element(self):
        generator_element = cElementTree.fromstring(
            """
            <generatoritem id="clipitem-29">
              <name>White</name>
              <enabled>TRUE</enabled>
              <duration>1035764</duration>
              <start>383</start>
              <end>432</end>
              <in>86313</in>
              <out>86362</out>
              <rate>
                <timebase>24</timebase>
                <ntsc>TRUE</ntsc>
              </rate>
              <effect>
                <name>Color</name>
                <effectid>Color</effectid>
                <effectcategory>Matte</effectcategory>
                <effecttype>generator</effecttype>
                <mediatype>video</mediatype>
                <parameter authoringApp="PremierePro">
                  <parameterid>fillcolor</parameterid>
                  <name>Color</name>
                  <value>
                    <alpha>0</alpha>
                    <red>255</red>
                    <green>255</green>
                    <blue>255</blue>
                  </value>
                </parameter>
              </effect>
            </generatoritem>
            """
        )
        parent_context_element = cElementTree.fromstring(
            """
            <track>
              <rate>
                <timebase>24</timebase>
                <ntsc>TRUE</ntsc>
              </rate>
            </track>
            """
        )

        context = self.adapter._Context(parent_context_element)

        # Make a parser
        parser = self.adapter.FCP7XMLParser(generator_element)

        clip, time_range = parser.item_and_timing_for_element(
            generator_element,
            head_transition=None,
            tail_transition=None,
            context=context,
        )

        self.assertEqual(clip.name, "White")

        expected_range = opentime.TimeRange(
            start_time=opentime.RationalTime(383, (24000 / 1001.0)),
            duration=opentime.RationalTime(49, (24000 / 1001.0)),
        )
        self.assertEqual(time_range, expected_range)

        expected_source_range = opentime.TimeRange(
            start_time=opentime.RationalTime(86313, (24000 / 1001.0)),
            duration=opentime.RationalTime(49, (24000 / 1001.0)),
        )
        self.assertEqual(clip.source_range, expected_source_range)

        ref = clip.media_reference
        self.assertTrue(
            isinstance(ref, schema.GeneratorReference)
        )
        self.assertEqual(ref.name, "Color")
        self.assertEqual(
            ref.metadata["fcp_xml"]["parameter"]["value"]["red"], "255"
        )

    def test_effect_from_filter_element(self):
        tree = cElementTree.parse(FILTER_XML_EXAMPLE_PATH)

        # Make a parser
        parser = self.adapter.FCP7XMLParser(tree)
        effect = parser.effect_from_filter_element(tree)

        self.assertEqual(effect.name, "Time Remap")

        # spot-check metadata
        effect_meta = effect.metadata["fcp_xml"]
        self.assertEqual(effect_meta["effectid"], "timeremap")
        self.assertEqual(len(effect_meta["parameter"]), 5)

    def test_transition_for_element(self):
        transition_element = cElementTree.fromstring(
            """
            <transitionitem>
              <start>538</start>
              <end>557</end>
              <alignment>end-black</alignment>
              <cutPointTicks>160876800000</cutPointTicks>
              <rate>
                <timebase>30</timebase>
                <ntsc>FALSE</ntsc>
              </rate>
              <effect>
                <name>Cross Dissolve</name>
                <effectid>Cross Dissolve</effectid>
                <effectcategory>Dissolve</effectcategory>
                <effecttype>transition</effecttype>
                <mediatype>video</mediatype>
                <wipecode>0</wipecode>
                <wipeaccuracy>100</wipeaccuracy>
                <startratio>0</startratio>
                <endratio>1</endratio>
                <reverse>FALSE</reverse>
              </effect>
            </transitionitem>
            """
        )

        track_element = cElementTree.fromstring(
            """
            <track>
                <rate>
                    <timebase>30</timebase>
                    <ntsc>FALSE</ntsc>
                </rate>
            </track>
            """
        )
        context = self.adapter._Context(track_element)

        parser = self.adapter.FCP7XMLParser(transition_element)
        transition = parser.transition_for_element(transition_element, context)

        self.assertEqual(transition.name, "Cross Dissolve")
        self.assertEqual(
            transition.transition_type,
            schema.TransitionTypes.SMPTE_Dissolve,
        )


class AdaptersFcp7XmlTest(unittest.TestCase, test_utils.OTIOAssertions):
    adapter = adapters.from_name('fcp_xml').module()

    def __init__(self, *args, **kwargs):
        super(AdaptersFcp7XmlTest, self).__init__(*args, **kwargs)
        self.maxDiff = None

    def test_build_empty_file(self):
        media_ref = schema.MissingReference(
            name="test_clip_name",
            available_range=opentime.TimeRange(
                opentime.RationalTime(820489, 24),
                opentime.RationalTime(2087, 24),
            ),
            metadata={
                "fcp_xml": {
                    "timecode": {
                        "rate": {"ntsc": "FALSE", "timebase": "24"},
                        "displayformat": "NDF",
                        "reel": {
                            "name": "test_reel_name",
                        },
                    }
                }
            },
        )

        file_element = self.adapter._build_empty_file(
            media_ref,
            media_ref.available_range.start_time,
            br_map={},
        )

        self.assertEqual(file_element.find("./name").text, "test_clip_name")
        self.assertEqual(file_element.find("./duration").text, "2087")

        rate_element = file_element.find("./rate")
        self.assertEqual(rate_element.find("./ntsc").text, "FALSE")
        self.assertEqual(rate_element.find("./timebase").text, "24")

        tc_element = file_element.find("./timecode")
        self.assertEqual(tc_element.find("./rate/ntsc").text, "FALSE")
        self.assertEqual(tc_element.find("./rate/timebase").text, "24")
        self.assertEqual(tc_element.find("./string").text, "09:29:47:01")
        self.assertEqual(tc_element.find("./reel/name").text, "test_reel_name")

    def test_read(self):
        timeline = adapters.read_from_file(FCP7_XML_EXAMPLE_PATH)

        self.assertTrue(timeline is not None)
        self.assertEqual(len(timeline.tracks), 8)

        video_tracks = [
            t for t in timeline.tracks
            if t.kind == schema.TrackKind.Video
        ]
        audio_tracks = [
            t for t in timeline.tracks
            if t.kind == schema.TrackKind.Audio
        ]

        self.assertEqual(len(video_tracks), 4)
        self.assertEqual(len(audio_tracks), 4)

        video_clip_names = (
            ("", 'sc01_sh010_anim.mov'),
            (
                "",
                'sc01_sh010_anim.mov',
                "",
                'sc01_sh020_anim.mov',
                'sc01_sh030_anim.mov',
                'Cross Dissolve',
                "",
                'sc01_sh010_anim'
            ),
            ("", 'test_title'),
            (
                "",
                'sc01_master_layerA_sh030_temp.mov',
                'Cross Dissolve',
                'sc01_sh010_anim.mov'
            )
        )

        for n, track in enumerate(video_tracks):
            self.assertTupleEqual(
                tuple(c.name for c in track),
                video_clip_names[n]
            )

        audio_clip_names = (
            ("", 'sc01_sh010_anim.mov', "", 'sc01_sh010_anim.mov'),
            ("", 'sc01_placeholder.wav', "", 'sc01_sh010_anim'),
            ("", 'track_08.wav'),
            ("", 'sc01_master_layerA_sh030_temp.mov', 'sc01_sh010_anim.mov')
        )

        for n, track in enumerate(audio_tracks):
            self.assertTupleEqual(
                tuple(c.name for c in track),
                audio_clip_names[n]
            )

        video_clip_durations = (
            ((536, 30.0), (100, 30.0)),
            (
                (13, 30.0),
                (100, 30.0),
                (52, 30.0),
                (157, 30.0),
                (235, 30.0),
                ((19, 30.0), (0, 30.0)),
                (79, 30.0),
                (320, 30.0)
            ),
            ((15, 30.0), (941, 30.0)),
            ((956, 30.0), (208, 30.0), ((12, 30.0), (13, 30.0)), (82, 30.0))
        )

        for t, track in enumerate(video_tracks):
            for c, clip in enumerate(track):
                if isinstance(clip, schema.Transition):
                    self.assertEqual(
                        clip.in_offset,
                        opentime.RationalTime(
                            *video_clip_durations[t][c][0]
                        )
                    )
                    self.assertEqual(
                        clip.out_offset,
                        opentime.RationalTime(
                            *video_clip_durations[t][c][1]
                        )
                    )
                else:
                    self.assertEqual(
                        clip.source_range.duration,
                        opentime.RationalTime(*video_clip_durations[t][c])
                    )

        audio_clip_durations = (
            ((13, 30.0), (100, 30.0), (423, 30.0), (100, 30.0), (423, 30.0)),
            (
                (335, 30.0),
                (170, 30.0),
                (131, 30.0),
                (294, 30.0),
                (34, 30.0),
                (124, 30.0)
            ),
            ((153, 30.0), (198, 30.0)),
            ((956, 30.0), (221, 30.0), (94, 30.0))
        )

        for t, track in enumerate(audio_tracks):
            for c, clip in enumerate(track):
                self.assertEqual(
                    clip.source_range.duration,
                    opentime.RationalTime(*audio_clip_durations[t][c])
                )

        timeline_marker_names = ('My MArker 1', 'dsf', "")

        for n, marker in enumerate(timeline.tracks.markers):
            self.assertEqual(marker.name, timeline_marker_names[n])

        timeline_marker_start_times = ((113, 30.0), (492, 30.0), (298, 30.0))

        for n, marker in enumerate(timeline.tracks.markers):
            self.assertEqual(
                marker.marked_range.start_time,
                opentime.RationalTime(*timeline_marker_start_times[n])
            )

        timeline_marker_comments = ('so, this happened', 'fsfsfs', None)

        for n, marker in enumerate(timeline.tracks.markers):
            self.assertEqual(
                marker.metadata.get('fcp_xml', {}).get('comment'),
                timeline_marker_comments[n]
            )

        clip_with_marker = video_tracks[1][4]
        clip_marker = clip_with_marker.markers[0]
        self.assertEqual(clip_marker.name, "")
        self.assertEqual(
            clip_marker.marked_range.start_time,
            opentime.RationalTime(73, 30.0)
        )
        self.assertEqual(
            clip_marker.metadata.get('fcp_xml', {}).get('comment'),
            None
        )

    def test_roundtrip_mem2disk2mem(self):
        timeline = schema.Timeline('test_timeline')

        RATE = 48.0

        video_reference = schema.ExternalReference(
            target_url="/var/tmp/test1.mov",
            available_range=opentime.TimeRange(
                opentime.RationalTime(value=100, rate=RATE),
                opentime.RationalTime(value=1000, rate=RATE)
            )
        )
        video_reference.name = "test_vid_one"
        audio_reference = schema.ExternalReference(
            target_url="/var/tmp/test1.wav",
            available_range=opentime.TimeRange(
                opentime.RationalTime(value=0, rate=RATE),
                opentime.RationalTime(value=1000, rate=RATE)
            ),
        )
        audio_reference.name = "test_wav_one"
        generator_reference = schema.GeneratorReference(
            name="Color",
            generator_kind="Color",
            metadata={
                "fcp_xml": {
                    "effectcategory": "Matte",
                    "effecttype": "generator",
                    "mediatype": "video",
                    "parameter": {
                        "@authoringApp": "PremierePro",
                        "parameterid": "fillcolor",
                        "name": "Color",
                        "value": {
                            "alpha": "0",
                            "red": "255",
                            "green": "255",
                            "blue": "255",
                        },
                    },
                },
            },
        )

        v0 = schema.Track(kind=schema.track.TrackKind.Video)
        v1 = schema.Track(kind=schema.track.TrackKind.Video)

        timeline.tracks.extend([v0, v1])

        a0 = schema.Track(kind=schema.track.TrackKind.Audio)

        timeline.tracks.append(a0)

        v0.extend(
            [
                schema.Clip(
                    name='test_clip1',
                    media_reference=video_reference,
                    source_range=opentime.TimeRange(
                        opentime.RationalTime(value=112, rate=RATE),
                        opentime.RationalTime(value=40, rate=RATE)
                    )
                ),
                schema.Gap(
                    source_range=opentime.TimeRange(
                        duration=opentime.RationalTime(
                            value=60,
                            rate=RATE
                        )
                    )
                ),
                schema.Clip(
                    name='test_clip2',
                    media_reference=video_reference,
                    source_range=opentime.TimeRange(
                        opentime.RationalTime(value=123, rate=RATE),
                        opentime.RationalTime(value=260, rate=RATE)
                    )
                ),
                schema.Clip(
                    name='test_generator_clip',
                    media_reference=generator_reference,
                    source_range=opentime.TimeRange(
                        opentime.RationalTime(value=292, rate=24.0),
                        opentime.RationalTime(value=183, rate=24.0)
                    )
                ),
            ]
        )

        v1.extend([
            schema.Gap(
                source_range=opentime.TimeRange(
                    duration=opentime.RationalTime(value=500, rate=RATE)
                )
            ),
            schema.Clip(
                name='test_clip3',
                media_reference=video_reference,
                source_range=opentime.TimeRange(
                    opentime.RationalTime(value=112, rate=RATE),
                    opentime.RationalTime(value=55, rate=RATE)
                )
            )
        ])

        a0.extend(
            [
                schema.Gap(
                    source_range=opentime.TimeRange(
                        duration=opentime.RationalTime(value=10, rate=RATE)
                    )
                ),
                schema.Clip(
                    name='test_clip4',
                    media_reference=audio_reference,
                    source_range=opentime.TimeRange(
                        opentime.RationalTime(value=152, rate=RATE),
                        opentime.RationalTime(value=248, rate=RATE)
                    ),
                )
            ]
        )

        timeline.tracks.markers.append(
            schema.Marker(
                name='test_timeline_marker',
                marked_range=opentime.TimeRange(
                    opentime.RationalTime(123, RATE)
                ),
                metadata={'fcp_xml': {'comment': 'my_comment'}}
            )
        )

        v1[1].markers.append(
            schema.Marker(
                name='test_clip_marker',
                marked_range=opentime.TimeRange(
                    opentime.RationalTime(125, RATE)
                ),
                metadata={'fcp_xml': {'comment': 'my_comment'}}
            )
        )

        # make sure that global_start_time.rate survives the round trip
        timeline.global_start_time = opentime.RationalTime(100, RATE)

        result = adapters.write_to_string(
            timeline,
            adapter_name='fcp_xml'
        )
        new_timeline = adapters.read_from_string(
            result,
            adapter_name='fcp_xml'
        )

        # Since FCP XML's "sequence" is a marriage of the timeline and the
        # main tracks stack, the tracks stack loses its name
        new_timeline.tracks.name = timeline.tracks.name

        self.assertEqual(new_timeline.name, 'test_timeline')

        # Before comparing, scrub ignorable metadata introduced in
        # serialization (things like unique ids minted by the adapter)
        # Since we seeded metadata for the generator, keep that metadata
        del new_timeline.metadata["fcp_xml"]
        for child in new_timeline.tracks.each_child():
            try:
                del child.metadata["fcp_xml"]
            except KeyError:
                pass

            try:
                is_generator = isinstance(
                    child.media_reference, schema.GeneratorReference
                )
                if not is_generator:
                    del child.media_reference.metadata["fcp_xml"]
            except (AttributeError, KeyError):
                pass

        self.assertJsonEqual(new_timeline, timeline)

    def test_roundtrip_disk2mem2disk(self):
        # somefile.xml -> OTIO
        timeline = adapters.read_from_file(FCP7_XML_EXAMPLE_PATH)
        tmp_path = tempfile.mkstemp(suffix=".xml", text=True)[1]

        # somefile.xml -> OTIO -> tempfile.xml
        adapters.write_to_file(timeline, tmp_path)

        # somefile.xml -> OTIO -> tempfile.xml -> OTIO
        result = adapters.read_from_file(tmp_path)

        # TODO: OTIO doesn't support linking items for the moment, so the
        # adapter reads links to the metadata, but doesn't write them.
        # See _dict_to_xml_tree for more information.
        def scrub_md_dicts(timeline):
            def scrub_displayformat(md_dict):
                for ignore_key in {"link"}:
                    try:
                        del md_dict[ignore_key]
                    except KeyError:
                        pass

                for value in list(md_dict.values()):
                    try:
                        value.items()
                        scrub_displayformat(value)
                    except AttributeError:
                        pass

            for child in timeline.tracks.each_child():
                scrub_displayformat(child.metadata)
                try:
                    scrub_displayformat(child.media_reference.metadata)
                except AttributeError:
                    pass

        # media reference bug, ensure that these match
        self.assertJsonEqual(
            result.tracks[0][1].media_reference,
            timeline.tracks[0][1].media_reference
        )

        scrub_md_dicts(result)
        scrub_md_dicts(timeline)

        self.assertJsonEqual(result, timeline)
        self.assertIsOTIOEquivalentTo(result, timeline)

        # But the xml text on disk is not identical because otio has a subset
        # of features to xml and we drop all the nle specific preferences.
        with open(FCP7_XML_EXAMPLE_PATH, "r") as original_file:
            with open(tmp_path, "r") as output_file:
                self.assertNotEqual(original_file.read(), output_file.read())

    def test_hiero_flavored_xml(self):
        timeline = adapters.read_from_file(HIERO_XML_PATH)
        self.assertTrue(len(timeline.tracks), 1)
        self.assertTrue(timeline.tracks[0].name == 'Video 1')

        clips = [c for c in timeline.tracks[0].each_clip()]
        self.assertTrue(len(clips), 2)

        self.assertTrue(clips[0].name == 'A160C005_171213_R0MN')
        self.assertTrue(clips[1].name == '/')

        self.assertTrue(
            isinstance(
                clips[0].media_reference,
                schema.ExternalReference
            )
        )

        self.assertTrue(
            isinstance(
                clips[1].media_reference,
                schema.MissingReference
            )
        )

        source_range = opentime.TimeRange(
            start_time=opentime.RationalTime(1101071, 24),
            duration=opentime.RationalTime(1055, 24)
        )
        self.assertTrue(clips[0].source_range == source_range)

        available_range = opentime.TimeRange(
            start_time=opentime.RationalTime(1101071, 24),
            duration=opentime.RationalTime(1055, 24)
        )
        self.assertTrue(clips[0].available_range() == available_range)

        clip_1_range = clips[1].available_range()
        self.assertEqual(
            clip_1_range,
            opentime.TimeRange(
                opentime.RationalTime(),
                opentime.RationalTime(1, 24),
            )
        )

        # Test serialization
        tmp_path = tempfile.mkstemp(suffix=".xml", text=True)[1]
        adapters.write_to_file(timeline, tmp_path)

        # Similar to the test_roundtrip_disk2mem2disk above
        # the track name element among others will not be present in a new xml.
        with open(HIERO_XML_PATH, "r") as original_file:
            with open(tmp_path, "r") as output_file:
                self.assertNotEqual(original_file.read(), output_file.read())

    def test_xml_with_empty_elements(self):
        timeline = adapters.read_from_file(EMPTY_ELEMENT_XML_PATH)

        # Spot-check the EDL, this one would throw exception on load before
        self.assertEqual(len(timeline.video_tracks()), 12)
        self.assertEqual(len(timeline.video_tracks()[0]), 34)

    def test_read_generators(self):
        timeline = adapters.read_from_file(GENERATOR_XML_EXAMPLE_PATH)

        video_track = timeline.tracks[0]
        audio_track = timeline.tracks[3]
        self.assertEqual(len(video_track), 6)
        self.assertEqual(len(audio_track), 3)

        # Check all video items are generators
        self.assertTrue(
            all(
                isinstance(item.media_reference, schema.GeneratorReference)
                for item in video_track
            )
        )

        # Check the video generator kinds
        self.assertEqual(
            [clip.media_reference.generator_kind for clip in video_track],
            ["Slug", "Slug", "Color", "Slug", "Slug", "GraphicAndType"],
        )

        # Check all non-gap audio items are generators
        self.assertTrue(
            all(
                isinstance(item.media_reference, schema.GeneratorReference)
                for item in video_track if not isinstance(item, schema.Gap)
            )
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Test Harness for the otio.opentime library."""

import opentimelineio as otio

import unittest
import copy


class TestTime(unittest.TestCase):

    def test_create(self):
        t_val = 30.2
        t = otio.opentime.RationalTime(t_val)
        self.assertIsNotNone(t)
        self.assertEqual(t.value, t_val)

        t_val = -30.2
        t = otio.opentime.RationalTime(t_val)
        self.assertIsNotNone(t)
        self.assertEqual(t.value, t_val)

        t = otio.opentime.RationalTime()
        self.assertEqual(t.value, 0)
        self.assertEqual(t.rate, 1.0)

    def test_equality(self):
        t1 = otio.opentime.RationalTime(30.2)
        self.assertEqual(t1, t1)
        t2 = otio.opentime.RationalTime(30.2)
        self.assertTrue(t1 is not t2)
        self.assertEqual(t1, t2)

    def test_inequality(self):
        t1 = otio.opentime.RationalTime(30.2)
        self.assertEqual(t1, t1)
        t2 = otio.opentime.RationalTime(33.2)
        self.assertTrue(t1 is not t2)
        self.assertNotEqual(t1, t2)
        t3 = otio.opentime.RationalTime(30.2)
        self.assertTrue(t1 is not t3)
        self.assertFalse(t1 != t3)

    def test_comparison(self):
        t1 = otio.opentime.RationalTime(15.2)
        t2 = otio.opentime.RationalTime(15.6)
        self.assertTrue(t1 < t2)
        self.assertTrue(t1 <= t2)
        self.assertFalse(t1 > t2)
        self.assertFalse(t1 >= t2)

        # Ensure the equality case of the comparisons works correctly
        t3 = otio.opentime.RationalTime(30.4, 2)
        self.assertTrue(t1 <= t3)
        self.assertTrue(t1 >= t3)
        self.assertTrue(t3 <= t1)
        self.assertTrue(t3 >= t1)

        # test implicit base conversion
        t2 = otio.opentime.RationalTime(15.6, 48)
        self.assertTrue(t1 > t2)
        self.assertTrue(t1 >= t2)
        self.assertFalse(t1 < t2)
        self.assertFalse(t1 <= t2)

    def test_copy(self):
        t1 = otio.opentime.RationalTime(18, 24)

        t2 = copy.copy(t1)
        self.assertEqual(t2, otio.opentime.RationalTime(18, 24))

    def test_deepcopy(self):
        t1 = otio.opentime.RationalTime(18, 24)

        t2 = copy.deepcopy(t1)
        self.assertEqual(t2, otio.opentime.RationalTime(18, 24))

    def test_base_conversion(self):
        # from a number
        t = otio.opentime.RationalTime(10, 24)
        with self.assertRaises(TypeError):
            t.rescaled_to("foo")
        self.assertEqual(t.rate, 24)
        t = t.rescaled_to(48)
        self.assertEqual(t.rate, 48)

        # from another RationalTime
        t = otio.opentime.RationalTime(10, 24)
        t2 = otio.opentime.RationalTime(20, 48)
        t = t.rescaled_to(t2)
        self.assertEqual(t.rate, t2.rate)

    def test_time_timecode_convert(self):
        timecode = "00:06:56:17"
        t = otio.opentime.from_timecode(timecode, 24)
        self.assertEqual(timecode, otio.opentime.to_timecode(t))

    def test_negative_timecode(self):
        with self.assertRaises(ValueError):
            otio.opentime.from_timecode('-01:00:13:13', 24)

    def test_bogus_timecode(self):
        with self.assertRaises(ValueError):
            otio.opentime.from_timecode('pink elephants', 13)

    def test_time_timecode_convert_bad_rate(self):
        with self.assertRaises(ValueError) as exception_manager:
            otio.opentime.from_timecode('01:00:13:24', 24)

        exc_message = str(exception_manager.exception)
        self.assertEqual(
            exc_message,
            "Frame rate mismatch.  Timecode '01:00:13:24' has frames beyond 23",
        )

    def test_timecode_24(self):
        timecode = "00:00:01:00"
        t = otio.opentime.RationalTime(value=24, rate=24)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

        timecode = "00:01:00:00"
        t = otio.opentime.RationalTime(value=24 * 60, rate=24)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

        timecode = "01:00:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60, rate=24)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

        timecode = "24:00:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60 * 24, rate=24)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

        timecode = "23:59:59:23"
        t = otio.opentime.RationalTime(value=24 * 60 * 60 * 24 - 1, rate=24)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

    def test_plus_equals(self):
        sum1 = otio.opentime.RationalTime()
        sum2 = otio.opentime.RationalTime()

        for i in range(10):
            incr = otio.opentime.RationalTime(i + 1, 24)
            sum1 += incr
            sum2 = sum2 + incr

        self.assertEqual(sum1, sum2)

    def test_time_timecode_zero(self):
        t = otio.opentime.RationalTime()
        timecode = "00:00:00:00"
        self.assertEqual(timecode, otio.opentime.to_timecode(t, 24))
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 24))

    def test_long_running_timecode_24(self):
        final_frame_number = 24 * 60 * 60 * 24 - 1
        final_time = otio.opentime.from_frames(final_frame_number, 24)
        self.assertEqual(
            otio.opentime.to_timecode(final_time),
            "23:59:59:23"
        )

        step_time = otio.opentime.RationalTime(value=1, rate=24)

        # fetching this test function from the c++ module directly
        cumulative_time = otio._opentime._testing.add_many(
            step_time,
            final_frame_number
        )
        self.assertEqual(cumulative_time, final_time)

        # Adding by a non-multiple of 24
        for fnum in range(1113, final_frame_number, 1113):
            rt = otio.opentime.from_frames(fnum, 24)
            tc = otio.opentime.to_timecode(rt)
            rt2 = otio.opentime.from_timecode(tc, 24)
            self.assertEqual(rt, rt2)
            self.assertEqual(tc, otio.opentime.to_timecode(rt2))

    def test_timecode_23976_fps(self):
        # This should behave exactly like 24 fps
        timecode = "00:00:01:00"
        t = otio.opentime.RationalTime(value=24, rate=23.976)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 23.976))

        timecode = "00:01:00:00"
        t = otio.opentime.RationalTime(value=24 * 60, rate=23.976)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 23.976))

        timecode = "01:00:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60, rate=23.976)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 23.976))

        timecode = "24:00:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60 * 24, rate=23.976)
        self.assertEqual(t, otio.opentime.from_timecode(timecode, 23.976))

        timecode = "23:59:59:23"
        t = otio.opentime.RationalTime(
            value=24 * 60 * 60 * 24 - 1,
            rate=(24000 / 1001.0)
        )
        self.assertEqual(
            t, otio.opentime.from_timecode(timecode, (24000 / 1001.0))
        )

    def test_converting_negative_values_to_timecode(self):
        t = otio.opentime.RationalTime(value=-1, rate=25)
        with self.assertRaises(ValueError):
            otio.opentime.to_timecode(t, 25)

    def test_dropframe_timecode_2997fps(self):
        """Test drop frame in action. Focused on minute roll overs

        We nominal_fps 30 for frame calculation
        For this frame rate we drop 2 frames per minute execpt every 10th.

        Compensation is calculated like this when below 10 minutes:
          (fps * seconds + frames - dropframes * (minutes - 1))
        Like this when not a whole 10 minute above 10 minutes:
          --minutes == minutes - 1
          (fps * seconds + frames - dropframes * (--minutes - --minutes / 10))
        And like this after that:
          (fps * seconds + frames - dropframes * (minutes - minutes / 10))
        """
        test_values = {
            'first_four_frames': [
                (0, '00:00:00;00'),
                (1, '00:00:00;01'),
                (2, '00:00:00;02'),
                (3, '00:00:00;03')
            ],

            'first_minute_rollover': [
                (30 * 59 + 29, '00:00:59;29'),
                (30 * 59 + 30, '00:01:00;02'),
                (30 * 59 + 31, '00:01:00;03'),
                (30 * 59 + 32, '00:01:00;04'),
                (30 * 59 + 33, '00:01:00;05')
            ],

            'fift_minute': [
                (30 * 299 + 29 - 2 * 4, '00:04:59;29'),
                (30 * 299 + 30 - 2 * 4, '00:05:00;02'),
                (30 * 299 + 31 - 2 * 4, '00:05:00;03'),
                (30 * 299 + 32 - 2 * 4, '00:05:00;04'),
                (30 * 299 + 33 - 2 * 4, '00:05:00;05')
            ],

            'seventh_minute': [
                (30 * 419 + 29 - 2 * 6, '00:06:59;29'),
                (30 * 419 + 30 - 2 * 6, '00:07:00;02'),
                (30 * 419 + 31 - 2 * 6, '00:07:00;03'),
                (30 * 419 + 32 - 2 * 6, '00:07:00;04'),
                (30 * 419 + 33 - 2 * 6, '00:07:00;05')
            ],

            'tenth_minute': [
                (30 * 599 + 29 - 2 * (10 - 10 // 10), '00:09:59;29'),
                (30 * 599 + 30 - 2 * (10 - 10 // 10), '00:10:00;00'),
                (30 * 599 + 31 - 2 * (10 - 10 // 10), '00:10:00;01'),
                (30 * 599 + 32 - 2 * (10 - 10 // 10), '00:10:00;02'),
                (30 * 599 + 33 - 2 * (10 - 10 // 10), '00:10:00;03')
            ],

            'second_hour': [
                (30 * 7199 + 29 - 2 * (120 - 120 // 10), '01:59:59;29'),
                (30 * 7199 + 30 - 2 * (120 - 120 // 10), '02:00:00;00'),
                (30 * 7199 + 31 - 2 * (120 - 120 // 10), '02:00:00;01'),
                (30 * 7199 + 32 - 2 * (120 - 120 // 10), '02:00:00;02'),
                (30 * 7199 + 33 - 2 * (120 - 120 // 10), '02:00:00;03')
            ],

            'second_and_a_half_hour': [
                (30 * 8999 + 29 - 2 * (150 - 150 // 10), '02:29:59;29'),
                (30 * 8999 + 30 - 2 * (150 - 150 // 10), '02:30:00;00'),
                (30 * 8999 + 31 - 2 * (150 - 150 // 10), '02:30:00;01'),
                (30 * 8999 + 32 - 2 * (150 - 150 // 10), '02:30:00;02'),
                (30 * 8999 + 33 - 2 * (150 - 150 // 10), '02:30:00;03')
            ],

            'tenth_hour': [
                (30 * 35999 + 29 - 2 * (600 - 600 // 10), '09:59:59;29'),
                (30 * 35999 + 30 - 2 * (600 - 600 // 10), '10:00:00;00'),
                (30 * 35999 + 31 - 2 * (600 - 600 // 10), '10:00:00;01'),
                (30 * 35999 + 32 - 2 * (600 - 600 // 10), '10:00:00;02'),
                (30 * 35999 + 33 - 2 * (600 - 600 // 10), '10:00:00;03')
            ],

            # Since 3 minutes < 10, we subtract 1 from 603 minutes
            'tenth_hour_third minute': [
                (30 * 36179 + 29 - 2 * (602 - 602 // 10), '10:02:59;29'),
                (30 * 36179 + 30 - 2 * (602 - 602 // 10), '10:03:00;02'),
                (30 * 36179 + 31 - 2 * (602 - 602 // 10), '10:03:00;03'),
                (30 * 36179 + 32 - 2 * (602 - 602 // 10), '10:03:00;04'),
                (30 * 36179 + 33 - 2 * (602 - 602 // 10), '10:03:00;05')
            ]
        }

        for time_key, time_values in test_values.items():
            for value, tc in time_values:
                t = otio.opentime.RationalTime(value, 29.97)
                self.assertEqual(
                    tc, otio.opentime.to_timecode(
                        t, rate=29.97, drop_frame=True
                    )
                )
                t1 = otio.opentime.from_timecode(tc, rate=29.97)
                self.assertEqual(t, t1)

    def test_timecode_ntsc_2997fps(self):
        frames = 1084319
        rate_float = (30000 / 1001.0)
        t = otio.opentime.RationalTime(frames, rate_float)

        dftc = otio.opentime.to_timecode(t, rate_float, drop_frame=True)
        self.assertEqual(dftc, '10:03:00;05')

        tc = otio.opentime.to_timecode(t, rate_float, drop_frame=False)
        self.assertEqual(tc, '10:02:23:29')

        # Detect DFTC from rate for backward compatability with old versions
        tc_auto = otio.opentime.to_timecode(t, rate_float)
        self.assertEqual(tc_auto, '10:03:00;05')

        invalid_df_rate = otio.opentime.RationalTime(30, (24000 / 1001.0))
        with self.assertRaises(ValueError):
            otio.opentime.to_timecode(
                invalid_df_rate, (24000 / 1001.0), drop_frame=True
            )

    def test_timecode_2997(self):
        ref_values = [
            (10789, '00:05:59:19', '00:05:59;29'),
            (10790, '00:05:59:20', '00:06:00;02'),
            (17981, '00:09:59:11', '00:09:59;29'),
            (17982, '00:09:59:12', '00:10:00;00'),
            (17983, '00:09:59:13', '00:10:00;01'),
            (17984, '00:09:59:14', '00:10:00;02'),
        ]

        for value, tc, dftc in ref_values:
            t = otio.opentime.RationalTime(value, 29.97)
            to_dftc = otio.opentime.to_timecode(t, rate=29.97, drop_frame=True)
            to_tc = otio.opentime.to_timecode(t, rate=29.97, drop_frame=False)
            to_auto_tc = otio.opentime.to_timecode(t, rate=29.97)

            # 29.97 should auto-detect dftc for backward compatability
            self.assertEqual(to_dftc, to_auto_tc)

            # check calculated against reference
            self.assertEqual(to_dftc, dftc)
            self.assertEqual(tc, to_tc)

            # Check they convert back
            t1 = otio.opentime.from_timecode(to_dftc, rate=29.97)
            self.assertEqual(t1, t)

            t2 = otio.opentime.from_timecode(to_tc, rate=29.97)
            self.assertEqual(t2, t)

    def test_faulty_formatted_timecode_24(self):
        """
        01:00:13;23 is drop-frame timecode, which only applies for
        NTSC rates (24000/1001, 30000/1001, etc). Such timecodes
        drop frames to compensate for the NTSC rates being slightly
        slower than whole frame rates (eg: 24 fps).
        It does not make sense to use drop-frame timecodes for
        non-NTSC rates.

        This is what we're testing here. When using 24 fps for the
        drop-frame timecode 01:00:13;23 we should get a ValueError
        mapping internally to the ErrorStatus
        'INVALID_RATE_FOR_DROP_FRAME_TIMECODE'.
        """
        with self.assertRaises(ValueError):
            otio.opentime.from_timecode('01:00:13;23', 24)

    def test_invalid_rate_to_timecode_functions(self):
        t = otio.opentime.RationalTime(100, 29.98)

        with self.assertRaises(ValueError):
            otio.opentime.to_timecode(t, 29.98)

        with self.assertRaises(ValueError):
            otio.opentime.to_timecode(t)

    def test_time_string_24(self):

        time_string = "00:00:00.041667"
        t = otio.opentime.RationalTime(value=1.0, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))
        self.assertEqual(time_obj.rate, 24)

        time_string = "00:00:01"
        t = otio.opentime.RationalTime(value=24, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "00:01:00"
        t = otio.opentime.RationalTime(value=24 * 60, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "01:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "24:00:00"
        t = otio.opentime.RationalTime(value=24 * 60 * 60 * 24, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "23:59:59.958333"
        t = otio.opentime.RationalTime(value=24 * 60 * 60 * 24 - 1, rate=24)
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

    def test_time_string_25(self):
        time_string = "00:00:01"
        t = otio.opentime.RationalTime(value=25, rate=25)
        time_obj = otio.opentime.from_time_string(time_string, 25)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "00:01:00"
        t = otio.opentime.RationalTime(value=25 * 60, rate=25)
        time_obj = otio.opentime.from_time_string(time_string, 25)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "01:00:00"
        t = otio.opentime.RationalTime(value=25 * 60 * 60, rate=25)
        time_obj = otio.opentime.from_time_string(time_string, 25)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "24:00:00"
        t = otio.opentime.RationalTime(value=25 * 60 * 60 * 24, rate=25)
        time_obj = otio.opentime.from_time_string(time_string, 25)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

        time_string = "23:59:59.92"
        t = otio.opentime.RationalTime(value=25 * 60 * 60 * 24 - 2, rate=25)
        time_obj = otio.opentime.from_time_string(time_string, 25)
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

    def test_time_time_string_negative_rational_time(self):
        """
        Negative rational time should return a valid time string
        with a '-' signage. (This is making it ffmpeg compatible)
        """

        baseline_time_string = "-00:00:01.0"
        rt = otio.opentime.RationalTime(-24, 24)
        time_string = otio.opentime.to_time_string(rt)
        self.assertEqual(baseline_time_string, time_string)

    def test_time_time_string_zero(self):
        t = otio.opentime.RationalTime()
        time_string = "00:00:00.0"
        time_obj = otio.opentime.from_time_string(time_string, 24)
        self.assertEqual(time_string, otio.opentime.to_time_string(t))
        self.assertTrue(t.almost_equal(time_obj, delta=0.001))

    def test_to_time_string_microseconds_starts_with_zero(self):
        # this number has a leading 0 in the fractional part when converted to
        # time string (ie 27.08333)
        rt = otio.opentime.RationalTime(2090, 24)
        self.assertEqual(
            str(rt),
            str(otio.opentime.from_time_string(otio.opentime.to_time_string(rt), 24))
        )

    def test_long_running_time_string_24(self):
        final_frame_number = 24 * 60 * 60 * 24 - 1
        final_time = otio.opentime.from_frames(final_frame_number, 24)
        self.assertEqual(
            otio.opentime.to_time_string(final_time),
            "23:59:59.958333"
        )

        step_time = otio.opentime.RationalTime(value=1, rate=24)
        cumulative_time = otio._opentime._testing.add_many(
            step_time,
            final_frame_number
        )

        self.assertTrue(cumulative_time.almost_equal(final_time, delta=0.001))

        # Adding by a non-multiple of 24
        for fnum in range(1113, final_frame_number, 1113):
            rt = otio.opentime.from_frames(fnum, 24)
            tc = otio.opentime.to_time_string(rt)
            rt2 = otio.opentime.from_time_string(tc, 24)
            self.assertEqual(rt, rt2)
            self.assertEqual(tc, otio.opentime.to_time_string(rt2))

    def test_time_string_23976_fps(self):
        # This list is rewritten from conversion into seconds of
        # test_timecode_23976_fps
        ref_values_23976 = [
            (1025, '00:00:01.708333'),
            (179900, '00:04:59.833333'),
            (180000, '00:05:00.0'),
            (360000, '00:10:00.0'),
            (720000, '00:20:00.0'),
            (1079300, '00:29:58.833333'),
            (1080000, '00:30:00.0'),
            (1080150, '00:30:00.25'),
            (1440000, '00:40:00.0'),
            (1800000, '00:50:00.0'),
            (1978750, '00:54:57.916666'),
            (1980000, '00:55:00.0'),
            (46700, '00:01:17.833333'),
            (225950, '00:06:16.583333'),
            (436400, '00:12:07.333333'),
            (703350, '00:19:32.25')
        ]
        for value, ts in ref_values_23976:
            t = otio.opentime.RationalTime(value, 600)
            self.assertEqual(ts, otio.opentime.to_time_string(t))
            # t1 = otio.opentime.from_time_string(ts, rate=23.976)
            # fails due to precision issues
            # self.assertEqual(t, t1)

    def test_time_to_string(self):
        t = otio.opentime.RationalTime(1.0, 2.0)
        self.assertEqual(str(t), "RationalTime(1, 2)")
        self.assertEqual(
            repr(t),
            "otio.opentime.RationalTime(value=1, rate=2)"
        )

    def test_frames_with_int_fps(self):
        for fps in (24, 30, 48, 60):
            t1 = otio.opentime.from_frames(101, fps)
            t2 = otio.opentime.RationalTime(101, fps)
            self.assertEqual(t1, t2)

    def test_frames_with_nonint_fps(self):
        for fps in (23.98, 29.97, 59.94):
            t1 = otio.opentime.from_frames(101, fps)
            t2 = otio.opentime.RationalTime(101, fps)
            self.assertEqual(t1, t2)

    def test_seconds(self):
        s1 = 1834
        t1 = otio.opentime.from_seconds(s1)
        self.assertEqual(t1.value, 1834)
        self.assertEqual(t1.rate, 1)
        t1_as_seconds = otio.opentime.to_seconds(t1)
        self.assertEqual(t1_as_seconds, s1)
        self.assertAlmostEqual(float(t1.value) / t1.rate, s1)

        s2 = 248474.345
        t2 = otio.opentime.from_seconds(s2)
        self.assertAlmostEqual(t2.value, s2)
        self.assertAlmostEqual(t2.rate, 1.0)
        t2_as_seconds = otio.opentime.to_seconds(t2)
        self.assertAlmostEqual(s2, t2_as_seconds)
        self.assertAlmostEqual(float(t2.value) / t2.rate, s2)

        v3 = 3459
        r3 = 24
        s3 = float(3459) / 24
        t3 = otio.opentime.RationalTime(v3, r3)
        t4 = otio.opentime.from_seconds(s3)
        self.assertAlmostEqual(otio.opentime.to_seconds(t3), s3)
        self.assertAlmostEqual(otio.opentime.to_seconds(t4), s3)

        t5 = otio.opentime.from_seconds(s3).rescaled_to(r3)
        t6 = otio.opentime.from_seconds(s3, r3)
        self.assertEqual(t5, t6)
        self.assertEqual(t6.rate, r3)

    def test_duration(self):
        start_time = otio.opentime.from_frames(100, 24)
        end = otio.opentime.from_frames(200, 24)
        duration = otio.opentime.duration_from_start_end_time(start_time, end)
        self.assertEqual(duration, otio.opentime.from_frames(100, 24))

        start_time = otio.opentime.from_frames(0, 1)
        end = otio.opentime.from_frames(200, 24)
        duration = otio.opentime.duration_from_start_end_time(start_time, end)
        self.assertEqual(duration, otio.opentime.from_frames(200, 24))

        start_time = otio.opentime.from_frames(100, 24)
        end = otio.opentime.from_frames(200, 24)
        duration = otio.opentime.duration_from_start_end_time_inclusive(start_time, end)
        self.assertEqual(duration, otio.opentime.from_frames(101, 24))

        start_time = otio.opentime.from_frames(0, 30)
        end = otio.opentime.from_frames(200, 24)
        duration = otio.opentime.duration_from_start_end_time_inclusive(start_time, end)
        self.assertEqual(duration, otio.opentime.from_frames(251, 30))

    def test_math(self):
        a = otio.opentime.from_frames(100, 24)
        gap = otio.opentime.from_frames(50, 24)
        b = otio.opentime.from_frames(150, 24)
        self.assertEqual(b - a, gap)
        self.assertEqual(a + gap, b)
        self.assertEqual(b - gap, a)

        with self.assertRaises(TypeError):
            b + "foo"

        a += gap
        self.assertEqual(a, b)

        a = otio.opentime.from_frames(100, 24)
        step = otio.opentime.from_frames(1, 24)
        for _ in range(50):
            a += step
        self.assertEqual(a, otio.opentime.from_frames(150, 24))

    def test_math_with_different_scales(self):
        a = otio.opentime.from_frames(100, 24)
        gap = otio.opentime.from_frames(100, 48)
        b = otio.opentime.from_frames(75, 12)
        self.assertEqual(b - a, gap.rescaled_to(24))
        self.assertEqual(a + gap, b.rescaled_to(48))
        gap2 = copy.copy(gap)
        gap2 += a
        self.assertEqual(gap2, a + gap)
        self.assertEqual(b - gap, a.rescaled_to(48))

    def test_duration_from_start_end_time(self):
        tend = otio.opentime.RationalTime(12, 25)

        tdur = otio.opentime.duration_from_start_end_time(
            start_time=otio.opentime.RationalTime(0, 25),
            end_time_exclusive=tend
        )

        self.assertEqual(tend, tdur)

    def test_subtract_with_different_rates(self):
        t1 = otio.opentime.RationalTime(12, 10)
        t2 = otio.opentime.RationalTime(12, 5)

        self.assertEqual((t1 - t2).value, -12)

    def test_incomparable_floats(self):
        t1 = otio.opentime.RationalTime(12, 10)
        with self.assertRaises(TypeError):
            t1 < -1

    def test_immutable(self):
        t1 = otio.opentime.RationalTime(12, 10)

        with self.assertRaises(AttributeError):
            t1.value = 12

    def test_passing_ndf_tc_at_df_rate(self):
        DF_TC = "01:00:02;05"
        NDF_TC = "00:59:58:17"
        frames = 107957

        tc1 = otio.opentime.to_timecode(
            otio.opentime.RationalTime(frames, 29.97)
        )
        self.assertEqual(tc1, DF_TC)

        tc2 = otio.opentime.to_timecode(
            otio.opentime.RationalTime(frames, 29.97),
            29.97,
            drop_frame=False
        )
        self.assertEqual(tc2, NDF_TC)

        t1 = otio.opentime.from_timecode(DF_TC, 29.97)
        self.assertEqual(t1.value, frames)

        t2 = otio.opentime.from_timecode(NDF_TC, 29.97)
        self.assertEqual(t2.value, frames)

    def test_nearest_valid_timecode_rate(self):
        invalid_valid_rates = (
            (23.97602397602397, 24000.0 / 1001.0),
            (23.97, 24000.0 / 1001.0),
            (23.976, 24000.0 / 1001.0),
            (23.98, 24000.0 / 1001.0),
            (29.97, 30000.0 / 1001.0),
            (59.94, 60000.0 / 1001.0),
        )

        for invalid_rate, nearest_valid_rate in invalid_valid_rates:
            self.assertTrue(
                otio.opentime.RationalTime.is_valid_timecode_rate(
                    nearest_valid_rate
                )
            )
            self.assertEqual(
                otio.opentime.RationalTime.nearest_valid_timecode_rate(
                    invalid_rate
                ),
                nearest_valid_rate,
            )


class TestTimeTransform(unittest.TestCase):

    def test_identity_transform(self):
        tstart = otio.opentime.RationalTime(12, 25)
        txform = otio.opentime.TimeTransform()
        self.assertEqual(tstart, txform.applied_to(tstart))

        tstart = otio.opentime.RationalTime(12, 25)
        txform = otio.opentime.TimeTransform(rate=50)
        self.assertEqual(24, txform.applied_to(tstart).value)

    def test_offset(self):
        tstart = otio.opentime.RationalTime(12, 25)
        toffset = otio.opentime.RationalTime(10, 25)
        txform = otio.opentime.TimeTransform(offset=toffset)
        self.assertEqual(tstart + toffset, txform.applied_to(tstart))

        tr = otio.opentime.TimeRange(tstart, tstart)
        self.assertEqual(
            txform.applied_to(tr),
            otio.opentime.TimeRange(tstart + toffset, tstart)
        )

    def test_scale(self):
        tstart = otio.opentime.RationalTime(12, 25)
        txform = otio.opentime.TimeTransform(scale=2)
        self.assertEqual(
            otio.opentime.RationalTime(24, 25),
            txform.applied_to(tstart)
        )

        tr = otio.opentime.TimeRange(tstart, tstart)
        tstart_scaled = otio.opentime.RationalTime(24, 25)
        self.assertEqual(
            txform.applied_to(tr),
            otio.opentime.TimeRange(tstart_scaled, tstart_scaled)
        )

    def test_rate(self):
        txform1 = otio.opentime.TimeTransform()
        txform2 = otio.opentime.TimeTransform(rate=50)
        self.assertEqual(txform2.rate, txform1.applied_to(txform2).rate)

    def test_string(self):
        tstart = otio.opentime.RationalTime(12.0, 25.0)
        txform = otio.opentime.TimeTransform(offset=tstart, scale=2.0)
        self.assertEqual(
            repr(txform),
            "otio.opentime.TimeTransform("
            "offset=otio.opentime.RationalTime("
            "value=12, "
            "rate=25"
            "), "
            "scale=2, "
            "rate=-1"
            ")"
        )

        self.assertEqual(
            str(txform),
            "TimeTransform(RationalTime(12, 25), 2, -1)"
        )

    def test_comparison(self):
        tstart = otio.opentime.RationalTime(12, 25)
        txform = otio.opentime.TimeTransform(offset=tstart, scale=2)
        tstart = otio.opentime.RationalTime(12, 25)
        txform2 = otio.opentime.TimeTransform(offset=tstart, scale=2)
        self.assertEqual(txform, txform2)
        self.assertFalse(txform != txform2)

        tstart = otio.opentime.RationalTime(23, 25)
        txform3 = otio.opentime.TimeTransform(offset=tstart, scale=2)
        self.assertNotEqual(txform, txform3)
        self.assertFalse(txform == txform3)

    def test_copy(self):
        tstart = otio.opentime.RationalTime(12, 25)
        t1 = otio.opentime.TimeTransform(tstart)

        t2 = copy.copy(t1)
        self.assertEqual(t1, t2)
        self.assertIsNot(t1, t2)
        self.assertEqual(t1.offset, t2.offset)
        # TimeTransform.__copy__ acts as a deep copy
        self.assertIsNot(t1.offset, t2.offset)

    def test_deepcopy(self):
        tstart = otio.opentime.RationalTime(12, 25)
        t1 = otio.opentime.TimeTransform(tstart)

        t2 = copy.deepcopy(t1)
        self.assertEqual(t1, t2)
        self.assertIsNot(t1, t2)
        self.assertEqual(t1.offset, t2.offset)
        # TimeTransform.__copy__ acts as a deep copy
        self.assertIsNot(t1.offset, t2.offset)


class TestTimeRange(unittest.TestCase):

    def test_create(self):
        tr = otio.opentime.TimeRange()
        blank = otio.opentime.RationalTime()
        self.assertEqual(tr.start_time, blank)
        self.assertEqual(tr.duration, blank)

        tr1 = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(10, 48)
        )
        self.assertEqual(tr1.start_time.rate, tr1.duration.rate)

        tr2 = otio.opentime.TimeRange(
            duration=otio.opentime.RationalTime(10, 48)
        )
        self.assertEqual(tr2.start_time.rate, tr2.duration.rate)

    def test_duration_validation(self):
        tr = otio.opentime.TimeRange()
        with self.assertRaises(AttributeError):
            setattr(tr, "duration", "foo")

    def test_extended_by(self):
        # base 25 is just for testing

        # range starts at 0 and has duration 0
        tr = otio.opentime.TimeRange(
            start_time=otio.opentime.RationalTime(0, 25)
        )
        with self.assertRaises(TypeError):
            tr.extended_by("foo")
        self.assertEqual(tr.duration, otio.opentime.RationalTime())

    def test_end_time(self):
        # test whole number duration
        rt_start = otio.opentime.RationalTime(1, 24)
        rt_dur = otio.opentime.RationalTime(5, 24)
        tr = otio.opentime.TimeRange(rt_start, rt_dur)
        self.assertEqual(tr.duration, rt_dur)
        self.assertEqual(tr.end_time_exclusive(), rt_start + rt_dur)
        self.assertEqual(
            tr.end_time_inclusive(),
            rt_start + rt_dur - otio.opentime.RationalTime(1, 24)
        )

        # test non-integer duration value
        rt_dur = otio.opentime.RationalTime(5.5, 24)
        tr = otio.opentime.TimeRange(rt_start, rt_dur)
        self.assertEqual(tr.end_time_exclusive(), rt_start + rt_dur)
        self.assertEqual(
            tr.end_time_inclusive(),
            otio.opentime.RationalTime(6, 24)
        )

    def test_repr(self):
        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(-1.0, 24.0),
            otio.opentime.RationalTime(6.0, 24.0)
        )
        self.assertEqual(
            repr(tr),
            "otio.opentime.TimeRange("
            "start_time=otio.opentime.RationalTime(value=-1, rate=24), "
            "duration=otio.opentime.RationalTime(value=6, rate=24))"
        )

    def test_compare(self):
        start_time1 = otio.opentime.RationalTime(18, 24)
        duration1 = otio.opentime.RationalTime(7, 24)
        tr1 = otio.opentime.TimeRange(start_time1, duration1)
        start_time2 = otio.opentime.RationalTime(18, 24)
        duration2 = otio.opentime.RationalTime(14, 48)
        tr2 = otio.opentime.TimeRange(start_time2, duration2)
        self.assertEqual(tr1, tr2)
        self.assertFalse(tr1 != tr2)

        start_time3 = otio.opentime.RationalTime(20, 24)
        duration3 = otio.opentime.RationalTime(3, 24)
        tr3 = otio.opentime.TimeRange(start_time3, duration3)
        self.assertNotEqual(tr1, tr3)
        self.assertFalse(tr1 == tr3)

    def test_copy(self):
        start_time1 = otio.opentime.RationalTime(18, 24)
        duration1 = otio.opentime.RationalTime(7, 24)
        tr1 = otio.opentime.TimeRange(start_time1, duration1)

        tr2 = copy.copy(tr1)
        self.assertEqual(
            tr2,
            otio.opentime.TimeRange(
                otio.opentime.RationalTime(18, 24),
                otio.opentime.RationalTime(7, 24),
            ),
        )

    def test_deepcopy(self):
        start_time1 = otio.opentime.RationalTime(18, 24)
        duration1 = otio.opentime.RationalTime(7, 24)
        tr1 = otio.opentime.TimeRange(start_time1, duration1)

        tr2 = copy.deepcopy(tr1)
        self.assertEqual(
            tr2,
            otio.opentime.TimeRange(
                otio.opentime.RationalTime(18, 24),
                otio.opentime.RationalTime(7, 24),
            ),
        )

    def test_clamped(self):
        test_point_min = otio.opentime.RationalTime(-2, 24)
        test_point_max = otio.opentime.RationalTime(6, 24)

        tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(-1, 24),
            otio.opentime.RationalTime(6, 24),
        )

        other_tr = otio.opentime.TimeRange(
            otio.opentime.RationalTime(-2, 24),
            otio.opentime.RationalTime(7, 24),
        )

        self.assertEqual(tr.clamped(test_point_min), tr.start_time)
        self.assertEqual(tr.clamped(test_point_max), tr.end_time_inclusive())

        self.assertEqual(tr.clamped(other_tr), tr)

    def test_overlaps_garbage(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        with self.assertRaises(TypeError):
            tr.overlaps("foo")

    def test_contains(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3.3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        with self.assertRaises(TypeError):
            tr.contains("foo")
        self.assertTrue(tr.contains(tstart))
        self.assertFalse(tr.contains(tstart + tdur))
        self.assertFalse(tr.contains(tstart - tdur))

        self.assertFalse(tr.contains(tr))

        tr_2 = otio.opentime.TimeRange(tstart - tdur, tdur)
        self.assertFalse(tr.contains(tr_2))
        self.assertFalse(tr_2.contains(tr))

    def test_overlaps_rationaltime(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.overlaps(otio.opentime.RationalTime(13, 25)))
        self.assertFalse(tr.overlaps(otio.opentime.RationalTime(1, 25)))

    def test_overlaps_timerange(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        tstart = otio.opentime.RationalTime(0, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(10, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(13, 25)
        tdur = otio.opentime.RationalTime(1, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(2, 25)
        tdur = otio.opentime.RationalTime(30, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(2, 50)
        tdur = otio.opentime.RationalTime(60, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(2, 50)
        tdur = otio.opentime.RationalTime(14, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(-100, 50)
        tdur = otio.opentime.RationalTime(400, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

        tstart = otio.opentime.RationalTime(100, 50)
        tdur = otio.opentime.RationalTime(400, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.overlaps(tr_t))

    def test_intersects_timerange(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        tstart = otio.opentime.RationalTime(0, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(10, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(10, 25)
        tdur = otio.opentime.RationalTime(2, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(14, 25)
        tdur = otio.opentime.RationalTime(2, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(15, 25)
        tdur = otio.opentime.RationalTime(2, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(13, 25)
        tdur = otio.opentime.RationalTime(1, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(2, 25)
        tdur = otio.opentime.RationalTime(30, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(2, 50)
        tdur = otio.opentime.RationalTime(60, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(2, 50)
        tdur = otio.opentime.RationalTime(14, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(-100, 50)
        tdur = otio.opentime.RationalTime(400, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.intersects(tr_t))

        tstart = otio.opentime.RationalTime(100, 50)
        tdur = otio.opentime.RationalTime(400, 50)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.intersects(tr_t))

    def test_before_timerange(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        tstart = otio.opentime.RationalTime(10, 25)
        tdur = otio.opentime.RationalTime(1.5, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertTrue(tr_t.before(tr))
        self.assertFalse(tr.before(tr_t))

        tdur = otio.opentime.RationalTime(12, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertFalse(tr_t.before(tr))

        self.assertFalse(tr.before(tr))

    def test_before_rationaltime(self):
        tafter = otio.opentime.RationalTime(15, 25)
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        self.assertFalse(tr.before(tafter))
        self.assertFalse(tr.before(tstart))

        tdur = otio.opentime.RationalTime(1.99, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        self.assertTrue(tr.before(tafter))

    def test_meets(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        tstart = otio.opentime.RationalTime(15, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.meets(tr_t))
        self.assertFalse(tr_t.meets(tr))

        tstart = otio.opentime.RationalTime(14.99, 25)
        tdur = otio.opentime.RationalTime(0, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr_t.meets(tr_t))

    def test_begins_timerange(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        tdur = otio.opentime.RationalTime(5, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.begins(tr_t))
        self.assertFalse(tr_t.begins(tr))
        self.assertFalse(tr.begins(tr))

        tdur = otio.opentime.RationalTime(0, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        self.assertTrue(tr.begins(tr_t))
        self.assertFalse(tr.begins(tr))

        tstart = otio.opentime.RationalTime(30, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertFalse(tr.begins(tr_t))

        tstart = otio.opentime.RationalTime(13, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        tdur = otio.opentime.RationalTime(3, 25)
        tstart = otio.opentime.RationalTime(12, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        self.assertFalse(tr_t.begins(tr))

    def test_begins_rationaltime(self):
        tend = otio.opentime.RationalTime(15, 25)
        tbefore = otio.opentime.RationalTime(11.9, 25)
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.begins(tstart))
        self.assertFalse(tr.begins(tend))
        self.assertFalse(tr.begins(tbefore))

    def test_finishes_timerange(self):
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)
        tstart = otio.opentime.RationalTime(13, 25)
        tdur = otio.opentime.RationalTime(2, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr_t.finishes(tr))
        self.assertFalse(tr.finishes(tr_t))
        self.assertFalse(tr.finishes(tr))

        tdur = otio.opentime.RationalTime(1, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertFalse(tr_t.finishes(tr))

        tstart = otio.opentime.RationalTime(30, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertFalse(tr_t.finishes(tr))

        tstart = otio.opentime.RationalTime(15, 25)
        tdur = otio.opentime.RationalTime(0, 25)
        tr_t = otio.opentime.TimeRange(tstart, tdur)
        self.assertTrue(tr_t.finishes(tr))

    def test_finishes_rationaltime(self):
        tafter = otio.opentime.RationalTime(16, 25)
        tend = otio.opentime.RationalTime(15, 25)
        tstart = otio.opentime.RationalTime(12, 25)
        tdur = otio.opentime.RationalTime(3, 25)
        tr = otio.opentime.TimeRange(tstart, tdur)

        self.assertTrue(tr.finishes(tend))
        self.assertFalse(tr.finishes(tstart))
        self.assertFalse(tr.finishes(tafter))

    def test_range_from_start_end_time(self):
        tstart = otio.opentime.RationalTime(0, 25)
        tend = otio.opentime.RationalTime(12, 25)

        tr = otio.opentime.range_from_start_end_time(
            start_time=tstart,
            end_time_exclusive=tend
        )

        self.assertEqual(tr.start_time, tstart)
        self.assertEqual(tr.duration, tend)

        self.assertEqual(tr.end_time_exclusive(), tend)
        self.assertEqual(
            tr.end_time_inclusive(),
            tend - otio.opentime.RationalTime(1, 25)
        )

        self.assertEqual(
            tr,
            otio.opentime.range_from_start_end_time(
                tr.start_time, tr.end_time_exclusive()
            )
        )

    def test_range_from_start_end_time_inclusive(self):
        tstart = otio.opentime.RationalTime(0, 25)
        tend = otio.opentime.RationalTime(12, 25)

        tr = otio.opentime.range_from_start_end_time_inclusive(
            start_time=tstart,
            end_time_inclusive=tend
        )

        self.assertEqual(tr.start_time, tstart)
        self.assertEqual(tr.duration, otio.opentime.RationalTime(13, 25))

        self.assertEqual(tr.end_time_inclusive(), tend)
        self.assertEqual(
            tr.end_time_inclusive(),
            otio.opentime.RationalTime(12, 25),
        )

        self.assertEqual(
            tr,
            otio.opentime.range_from_start_end_time_inclusive(
                tr.start_time, tr.end_time_inclusive()
            )
        )

    def test_adjacent_timeranges(self):
        d1 = 0.3
        d2 = 0.4
        r1 = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(d1, 1)
        )
        r2 = otio.opentime.TimeRange(
            r1.end_time_exclusive(),
            otio.opentime.RationalTime(d2, 1)
        )
        full = otio.opentime.TimeRange(
            otio.opentime.RationalTime(0, 1),
            otio.opentime.RationalTime(d1 + d2, 1)
        )
        self.assertFalse(r1.overlaps(r2))
        self.assertEqual(r1.extended_by(r2), full)

    def test_distant_timeranges(self):
        start = 0.1
        d1 = 0.3
        gap = 1.7
        d2 = 0.4
        r1 = otio.opentime.TimeRange(
            otio.opentime.RationalTime(start, 1),
            otio.opentime.RationalTime(d1, 1)
        )
        r2 = otio.opentime.TimeRange(
            otio.opentime.RationalTime(start + gap + d1, 1),
            otio.opentime.RationalTime(d2, 1)
        )
        full = otio.opentime.TimeRange(
            otio.opentime.RationalTime(start, 1),
            otio.opentime.RationalTime(d1 + gap + d2, 1)
        )
        self.assertFalse(r1.overlaps(r2))
        self.assertEqual(full, r1.extended_by(r2))
        self.assertEqual(full, r2.extended_by(r1))

    def test_to_timecode_mixed_rates(self):
        timecode = "00:06:56:17"
        t = otio.opentime.from_timecode(timecode, 24)
        self.assertEqual(timecode, otio.opentime.to_timecode(t))
        self.assertEqual(timecode, otio.opentime.to_timecode(t, 24))
        self.assertNotEqual(timecode, otio.opentime.to_timecode(t, 12))

        time1 = otio.opentime.RationalTime(24.0, 24.0)
        time2 = otio.opentime.RationalTime(1.0, 1.0)
        self.assertEqual(
            otio.opentime.to_timecode(time1, 24.0),
            otio.opentime.to_timecode(time2, 24.0)
        )

    def test_to_frames_mixed_rates(self):
        frame = 100
        t = otio.opentime.from_frames(frame, 24)
        self.assertEqual(frame, otio.opentime.to_frames(t))
        self.assertEqual(frame, otio.opentime.to_frames(t, 24))
        self.assertNotEqual(frame, otio.opentime.to_frames(t, 12))


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import unittest
import os
import sys
import subprocess

from opentimelineio.console import (
    autogen_serialized_datamodel as asd,
    autogen_plugin_documentation as apd,
    autogen_version_map as avm
)


@unittest.skipIf(
    os.environ.get("OTIO_DISABLE_SERIALIZED_SCHEMA_TEST"),
    "Serialized schema test disabled because "
    "$OTIO_DISABLE_SERIALIZED_SCHEMA_TEST is set to something other than ''"
)
class SerializedSchemaTester(unittest.TestCase):
    def test_serialized_schema(self):
        """Test if the schema has changed since last time the serialized schema
        documentation was generated.
        """

        pt = os.path.dirname(os.path.dirname(__file__))
        fp = os.path.join(pt, "docs", "tutorials", "otio-serialized-schema.md")
        with open(fp) as fi:
            baseline_text = fi.read()

        test_text, _ = asd.generate_and_write_documentation()

        self.maxDiff = None
        self.longMessage = True
        self.assertMultiLineEqual(
            baseline_text,
            test_text,
            "\n The schema has changed and the autogenerated documentation in {}"
            " needs to be updated.  run: `make doc-model-update`".format(fp)
        )


@unittest.skipIf(
    os.environ.get("OTIO_DISABLE_SERIALIZED_SCHEMA_TEST"),
    "Plugin documentation test disabled because "
    "$OTIO_DISABLE_SERIALIZED_SCHEMA_TEST is set to something other than ''"
)
class PluginDocumentationTester(unittest.TestCase):
    def test_plugin_documentation(self):
        """Verify that the plugin manifest matches what is checked into the
        documentation.
        """

        pt = os.path.dirname(os.path.dirname(__file__))
        fp = os.path.join(pt, "docs", "tutorials", "otio-plugins.md")
        with open(fp) as fi:
            baseline_text = fi.read()

        test_text = apd.generate_and_write_documentation_plugins(
            public_only=True,
            sanitized_paths=True
        )

        self.maxDiff = None
        self.longMessage = True
        self.assertMultiLineEqual(
            baseline_text,
            test_text,
            "\n The schema has changed and the autogenerated documentation in {}"
            " needs to be updated.  run: `make doc-plugins-update`".format(fp)
        )


@unittest.skipIf(
    os.environ.get("OTIO_DISABLE_SERIALIZED_SCHEMA_TEST"),
    "CORE_VERSION_MAP generation test disabled because "
    "$OTIO_DISABLE_SERIALIZED_SCHEMA_TEST is set to something other than ''"
)
class CoreVersionMapGenerationTester(unittest.TestCase):
    def test_core_version_map_generator(self):
        """Verify the current CORE_VERSION_MAP matches the checked in one."""

        pt = os.path.dirname(os.path.dirname(__file__))
        root = os.path.join(pt, "src", "opentimelineio")
        template_fp = os.path.join(root, "CORE_VERSION_MAP.last.cpp")
        target_fp = os.path.join(root, "CORE_VERSION_MAP.cpp")

        with open(target_fp) as fi:
            # sanitize line endings and remove empty lines for cross-windows
            # /*nix consistent behavior
            baseline_text = "\n".join(
                ln
                for ln in fi.read().splitlines()
                if ln
            )

        proc = subprocess.Popen(
            [
                sys.executable,
                avm.__file__,
                "-i",
                template_fp,
                "-d",
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, _ = proc.communicate()

        test_text = stdout.decode("utf-8")[:-1]

        # sanitize the line endings
        test_text = "\n".join(
            ln
            for ln in test_text.splitlines()
            if ln
        )

        self.maxDiff = None
        self.longMessage = True
        self.assertMultiLineEqual(
            baseline_text,
            test_text,
            "\n The CORE_VERSION_MAP has changed and the autogenerated one in"
            " {} needs to be updated. run: `make version-map-update`".format(
                target_fp
            )
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import unittest

import opentimelineio as otio
import opentimelineio.test_utils as otio_test_utils

has_undefined_schema = """
{
    "OTIO_SCHEMA": "Clip.1",
    "effects": [],
    "markers": [],
    "media_reference": {
        "OTIO_SCHEMA": "ExternalReference.1",
        "available_range": {
            "OTIO_SCHEMA": "TimeRange.1",
            "duration": {
                "OTIO_SCHEMA": "RationalTime.1",
                "rate": 24,
                "value": 140
            },
            "start_time": {
                "OTIO_SCHEMA": "RationalTime.1",
                "rate": 24,
                "value": 91
            }
        },
        "metadata": {
            "stuff": {
                "OTIO_SCHEMA": "MyOwnDangSchema.3",
                "some_data": 895,
                "howlongami": {
                     "OTIO_SCHEMA": "RationalTime.1",
                      "rate": 30,
                      "value": 100
                   }
            }
        },
        "name": null,
        "target_url": "/usr/tmp/some_media.mov"
    },
    "metadata": {},
    "name": null,
    "source_range": null
}
"""


class UnknownSchemaTests(unittest.TestCase, otio_test_utils.OTIOAssertions):
    def setUp(self):
        # make an OTIO data structure containing an undefined schema object
        self.orig = otio.adapters.otio_json.read_from_string(has_undefined_schema)

    def test_serialize_deserialize(self):
        serialized = otio.adapters.otio_json.write_to_string(self.orig)
        test_otio = otio.adapters.otio_json.read_from_string(serialized)

        self.assertIsOTIOEquivalentTo(self.orig, test_otio)

    def test_is_unknown_schema(self):
        self.assertFalse(self.orig.is_unknown_schema)
        unknown = self.orig.media_reference.metadata["stuff"]
        self.assertTrue(unknown.is_unknown_schema)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unit tests for the maya sequencer adapter"""

import os
import tempfile
import unittest

import opentimelineio as otio

SAMPLE_DATA_DIR = os.path.join(os.path.dirname(__file__), "sample_data")
SCREENING_EXAMPLE_PATH = os.path.join(SAMPLE_DATA_DIR, "screening_example.edl")
BASELINE_PATH = os.path.join(SAMPLE_DATA_DIR, "screening_example.ma")
SETATTR_TO_CHECK = (".ef", ".sf", ".sn", ".se", ".ssf")


def filter_maya_file(contents):
    return '\n'.join(
        line for line in contents.split('\n')
        if (
            line.strip().startswith('setAttr') and
            any(a in line for a in SETATTR_TO_CHECK) or
            (
                not line.startswith('//') and
                not line.startswith('requires') and
                not line.startswith('fileInfo') and
                not line.startswith('currentUnit') and
                not line.strip().startswith('rename') and
                not line.strip().startswith('select') and
                not line.strip().startswith('setAttr') and
                not line.strip().startswith('0') and
                not line.strip().startswith('1')
            )
        )
    )


@unittest.skipIf(
    "OTIO_MAYA_PYTHON_BIN" not in os.environ,
    "OTIO_MAYA_PYTHON_BIN not set, required for the maya adapter"
)
class MayaSequencerAdapterWriteTest(unittest.TestCase):
    def test_basic_maya_sequencer_write(self):
        self.maxDiff = None
        timeline = otio.adapters.read_from_file(SCREENING_EXAMPLE_PATH)
        tmp_path = tempfile.mkstemp(suffix=".ma", text=True)[1]

        otio.adapters.write_to_file(timeline, tmp_path)
        self.assertTrue(os.path.exists(tmp_path))

        with open(tmp_path) as fo:
            test_data = fo.read()

        with open(BASELINE_PATH) as fo:
            baseline_data = fo.read()

        self.assertMultiLineEqual(
            filter_maya_file(baseline_data),
            filter_maya_file(test_data)
        )


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

import os
import unittest

from tests import baseline_reader

import opentimelineio as otio
from tests import utils

LINKER_PATH = "media_linker_example"


class TestPluginMediaLinker(unittest.TestCase):
    def setUp(self):
        self.bak = otio.plugins.ActiveManifest()
        self.man = utils.create_manifest()
        otio.plugins.manifest._MANIFEST = self.man
        self.jsn = baseline_reader.json_baseline_as_string(LINKER_PATH)
        self.mln = otio.adapters.otio_json.read_from_string(self.jsn)
        self.mln._json_path = os.path.join(
            baseline_reader.MODPATH,
            "baselines",
            LINKER_PATH
        )

    def tearDown(self):
        otio.plugins.manifest._MANIFEST = self.bak
        utils.remove_manifest(self.man)

    def test_plugin_adapter(self):
        self.assertEqual(self.mln.name, "example")
        self.assertEqual(self.mln.execution_scope, "in process")
        self.assertEqual(self.mln.filepath, "example.py")

    def test_load_adapter_module(self):

        target = os.path.join(
            baseline_reader.MODPATH,
            "baselines",
            "example.py"
        )

        self.assertEqual(self.mln.module_abs_path(), target)
        self.assertTrue(hasattr(self.mln.module(), "link_media_reference"))

    def test_run_linker(self):
        cl = otio.schema.Clip(name="foo")

        linked_mr = self.mln.link_media_reference(cl, {"extra_data": True})
        self.assertIsInstance(linked_mr, otio.schema.MissingReference)
        self.assertEqual(linked_mr.name, cl.name + "_tweaked")
        self.assertEqual(linked_mr.metadata.get("extra_data"), True)

    def test_serialize(self):

        self.assertEqual(
            str(self.mln),
            "MediaLinker({}, {}, {})".format(
                repr(self.mln.name),
                repr(self.mln.execution_scope),
                repr(self.mln.filepath)
            )
        )
        self.assertEqual(
            repr(self.mln),
            "otio.media_linker.MediaLinker("
            "name={}, "
            "execution_scope={}, "
            "filepath={}"
            ")".format(
                repr(self.mln.name),
                repr(self.mln.execution_scope),
                repr(self.mln.filepath)
            )
        )

    def test_available_media_linker_names(self):
        # for now just assert that it returns a non-empty list
        self.assertTrue(otio.media_linker.available_media_linker_names())

    def test_default_media_linker(self):
        os.environ['OTIO_DEFAULT_MEDIA_LINKER'] = 'foo'
        self.assertEqual(otio.media_linker.default_media_linker(), 'foo')
        with self.assertRaises(otio.exceptions.NoDefaultMediaLinkerError):
            del os.environ['OTIO_DEFAULT_MEDIA_LINKER']
            otio.media_linker.default_media_linker()

    def test_from_name_fail(self):
        with self.assertRaises(otio.exceptions.NotSupportedError):
            otio.media_linker.from_name("should not exist")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Example OTIO script that generates an OTIO from a single quicktime by using
ffprobe to detect shot breaks.
"""

import subprocess
import re
import os
import argparse

import opentimelineio as otio


class ShotDetectError(Exception):
    pass


class FFProbeFailedError(ShotDetectError):
    pass


"""
@NOTE EXAMPLE LINE from ffprobe:
media_type=video|stream_index=0|key_frame=1|pkt_pts=385875|pkt_pts_time=128.753754|pkt_dts=385875|pkt_dts_time=128.753754|best_effort_timestamp=385875|best_effort_timestamp_time=128.753754|pkt_duration=125|pkt_duration_time=0.041708|pkt_pos=160853826|pkt_size=4608000|width=1920|height=800|pix_fmt=rgb24|sample_aspect_ratio=1:1|pict_type=I|coded_picture_number=0|display_picture_number=0|interlaced_frame=0|top_field_first=0|repeat_pict=0|tag:lavfi.scene_score=0.421988
"""


def parse_args():
    """ parse arguments out of sys.argv """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'filepath',
        type=str,
        nargs='+',
        help='Files to look for splits in.'
    )
    parser.add_argument(
        '-r',
        '--reference-clip',
        action='store_true',
        default=False,
        help="instead of detecting shots, write a timeline with a single clip "
        "referencing the file."
    )
    parser.add_argument(
        '-d',
        '--dryrun',
        action="store_true",
        default=False,
        help="Print ffprobe command instead of running it."
    )
    return parser.parse_args()


_MEDIA_RANGE_MAP = {}


def _media_start_end_of(media_path, fps):
    if media_path in _MEDIA_RANGE_MAP:
        return _MEDIA_RANGE_MAP[media_path]

    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "format=duration",
        "-of",
        "default=noprint_wrappers=1:nokey=1",
        media_path
    ]
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode != 0:
        raise FFProbeFailedError(
            "FFProbe Failed with error: {0}, output: {1}".format(
                err, out
            )
        )

    available_range = otio.opentime.TimeRange(
        otio.opentime.RationalTime(0, 1).rescaled_to(fps),
        otio.opentime.RationalTime(float(out), 1).rescaled_to(fps)
    )

    _MEDIA_RANGE_MAP[media_path] = available_range

    return available_range


def _timeline_with_single_clip(name, full_path, dryrun=False):
    timeline = otio.schema.Timeline()
    timeline.name = name
    track = otio.schema.Track()
    track.name = name
    timeline.tracks.append(track)

    fps = _ffprobe_fps(name, full_path, dryrun)
    available_range = _media_start_end_of(full_path, fps)

    media_reference = otio.schema.ExternalReference(
        target_url="file://" + full_path,
        available_range=available_range
    )

    if dryrun:
        return

    clip = otio.schema.Clip(name=name)
    clip.media_reference = media_reference
    track.append(clip)
    return timeline


def _timeline_with_breaks(name, full_path, dryrun=False):

    timeline = otio.schema.Timeline()
    timeline.name = name
    track = otio.schema.Track()
    track.name = name
    timeline.tracks.append(track)

    fps = _ffprobe_fps(name, full_path, dryrun)

    out, _ = _ffprobe_output(name, full_path, dryrun)

    if dryrun:
        return ("", "")

    # saving time in base 1.0 to demonstrate that you can
    playhead = otio.opentime.RationalTime(0, 1.0)
    shot_index = 1

    for line in out.split("\n"):
        info = dict(re.findall(r'(\w+)=(\d+\.\d*)', line))

        end_time_in_seconds = info.get('pkt_pts_time')

        if end_time_in_seconds is None:
            continue

        start_time = playhead.rescaled_to(fps)

        # Note: if you wanted to snap to a particular frame rate, you would do
        # it here.
        end_time_exclusive = otio.opentime.RationalTime(
            float(end_time_in_seconds),
            1.0
        ).rescaled_to(fps)

        clip = otio.schema.Clip()
        clip.name = "shot {0}".format(shot_index)
        clip.source_range = otio.opentime.range_from_start_end_time(
            start_time,
            end_time_exclusive
        )

        available_range = _media_start_end_of(full_path, fps)

        clip.media_reference = otio.schema.ExternalReference(
            target_url="file://" + full_path,
            available_range=available_range
        )
        track.append(clip)

        playhead = end_time_exclusive
        shot_index += 1

    return timeline


def _verify_ffprobe():
    cmd = [
        "ffprobe",
        "-h"
    ]
    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
    except OSError:
        raise FFProbeFailedError(
            "Unable to run ffprobe command line tool. It might not be in your "
            "path, or you might need to get it from https://www.ffmpeg.org"
        )
    out, err = proc.communicate()
    if proc.returncode != 0:
        raise FFProbeFailedError(
            "FFProbe Failed with error: {0}, output: {1}".format(
                err, out
            )
        )


def _ffprobe_fps(name, full_path, dryrun=False):
    """Parse the framerate from the file using ffprobe."""

    _, err = _ffprobe_output(
        name,
        full_path,
        dryrun,
        arguments=["{0}".format(full_path)],
        message="framerate"
    )

    if dryrun:
        return 1.0

    try:
        # python3
        # if its bytes (python3+)
        err_str = err.decode("utf-8")
    except AttributeError:
        # if its a string (python <3)
        err_str = err

    for line in err_str.split('\n'):
        if not ("Stream" in line and "Video" in line):
            continue

        bits = line.split(",")
        for bit in bits:
            if "fps" not in bit:
                continue
            return float([b for b in bit.split(" ") if b][0])

    # fallback FPS is just 1.0 -- seconds
    return 1.0


def _ffprobe_output(
    name,
    full_path,
    dryrun=False,
    arguments=None,
    message="shot breaks"
):
    """ Run ffprobe and return resulting output """

    arguments = arguments or [
        "-show_frames",
        "-of",
        "compact=p=0",
        "-f",
        "lavfi",
        "movie={0},select=gt(scene\\,.1)".format(full_path)
    ]

    if message:
        print("Scanning {0} for {1}...".format(name, message))

    cmd = ["ffprobe"] + arguments

    if dryrun:
        print(" ".join(cmd))
        return ("", "")

    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode != 0:
        raise FFProbeFailedError(
            "FFProbe Failed with error: {0}, output: {1}".format(
                err, out
            )
        )

    return out, err


def main():
    args = parse_args()

    # make sure that ffprobe exists before continuing
    _verify_ffprobe()

    for full_path in args.filepath:
        name = os.path.basename(full_path)

        if args.reference_clip:
            new_tl = _timeline_with_single_clip(name, full_path, args.dryrun)
        else:
            new_tl = _timeline_with_breaks(name, full_path, args.dryrun)

        if args.dryrun:
            continue

        # @TODO: this should be an argument
        otio_filename = os.path.splitext(name)[0] + ".otio"
        otio.adapters.write_to_file(new_tl, otio_filename)
        print(
            "SAVED: {0} with {1} clips.".format(
                otio_filename,
                len(new_tl.tracks[0])
            )
        )


if __name__ == '__main__':
    main()
